#   Copyright 2018 Red Hat, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import mock

from osc_lib.tests import utils

from tripleoclient import constants
from tripleoclient.tests import fakes
from tripleoclient.v1 import undercloud_backup


class TestUndercloudBackup(utils.TestCommand):

    def setUp(self):
        super(TestUndercloudBackup, self).setUp()

        # Get the command object to test
        app_args = mock.Mock()
        app_args.verbose_level = 1
        self.app.options = fakes.FakeOptions()
        self.cmd = undercloud_backup.BackupUndercloud(self.app, app_args)
        self.app.client_manager.workflow_engine = mock.Mock()
        self.workflow = self.app.client_manager.workflow_engine

    @mock.patch('tripleoclient.utils.run_ansible_playbook',
                autospec=True)
    def test_undercloud_backup_legacy_withargs(self, mock_playbook):
        arglist = [
            '--add-path',
            '/tmp/foo.yaml',
            '--add-path',
            '/tmp/bar.yaml'
        ]
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)
        mock_playbook.assert_called_once_with(
            workdir=mock.ANY,
            playbook=mock.ANY,
            inventory=mock.ANY,
            skip_tags=None,
            playbook_dir=constants.ANSIBLE_TRIPLEO_PLAYBOOKS,
            verbosity=3,
            extra_vars={'sources_path':
                        '/home/stack/,/tmp/bar.yaml,/tmp/foo.yaml'})

    @mock.patch('tripleoclient.utils.run_ansible_playbook',
                autospec=True)
    def test_undercloud_backup_withargs_remove(self, mock_playbook):
        arglist = [
            '--add-path',
            '/tmp/foo.yaml',
            '--exclude-path',
            '/tmp/bar.yaml',
            '--exclude-path',
            '/home/stack/',
            '--add-path',
            '/tmp/bar.yaml'
        ]
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)
        mock_playbook.assert_called_once_with(
            workdir=mock.ANY,
            playbook=mock.ANY,
            inventory=mock.ANY,
            skip_tags=None,
            verbosity=3,
            playbook_dir=constants.ANSIBLE_TRIPLEO_PLAYBOOKS,
            extra_vars={'sources_path':
                        '/tmp/foo.yaml'})

    @mock.patch('tripleoclient.utils.run_ansible_playbook',
                autospec=True)
    def test_undercloud_backup_withargs_remove_double(self, mock_playbook):
        arglist = [
            '--add-path',
            '/tmp/foo.yaml',
            '--add-path',
            '/tmp/bar.yaml',
            '--exclude-path',
            '/tmp/foo.yaml',
            '--exclude-path',
            '/tmp/foo.yaml'
        ]
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)
        mock_playbook.assert_called_once_with(
            workdir=mock.ANY,
            playbook=mock.ANY,
            inventory=mock.ANY,
            skip_tags=None,
            playbook_dir=constants.ANSIBLE_TRIPLEO_PLAYBOOKS,
            verbosity=3,
            extra_vars={'sources_path':
                        '/home/stack/,/tmp/bar.yaml'})

    @mock.patch('tripleoclient.utils.run_ansible_playbook',
                autospec=True)
    def test_undercloud_backup_withargs_remove_unex(self, mock_playbook):
        arglist = [
            '--add-path',
            '/tmp/foo.yaml',
            '--exclude-path',
            '/tmp/non-existing-path.yaml'
        ]
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)
        mock_playbook.assert_called_once_with(
            workdir=mock.ANY,
            playbook=mock.ANY,
            inventory=mock.ANY,
            skip_tags=None,
            playbook_dir=constants.ANSIBLE_TRIPLEO_PLAYBOOKS,
            verbosity=3,
            extra_vars={'sources_path':
                        '/home/stack/,/tmp/foo.yaml'})

    @mock.patch('tripleoclient.utils.run_ansible_playbook',
                autospec=True)
    def test_undercloud_backup_noargs(self, mock_playbook):
        arglist = []
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)
        mock_playbook.assert_called_once_with(
            workdir=mock.ANY,
            playbook='cli-undercloud-backup.yaml',
            inventory=parsed_args.inventory,
            skip_tags=None,
            playbook_dir=constants.ANSIBLE_TRIPLEO_PLAYBOOKS,
            verbosity=3,
            extra_vars=None
        )

    @mock.patch('tripleoclient.utils.run_ansible_playbook',
                autospec=True)
    def test_undercloud_backup_init(self, mock_playbook):
        arglist = [
            '--init'
        ]
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)
        mock_playbook.assert_called_once_with(
            workdir=mock.ANY,
            playbook='prepare-undercloud-backup.yaml',
            inventory=parsed_args.inventory,
            skip_tags='bar_create_recover_image',
            playbook_dir=constants.ANSIBLE_TRIPLEO_PLAYBOOKS,
            verbosity=3,
            extra_vars=None
            )

    @mock.patch('tripleoclient.utils.run_ansible_playbook',
                autospec=True)
    def test_undercloud_backup_init_with_inventory(self, mock_playbook):
        arglist = [
            '--init',
            '--inventory',
            '/tmp/test_inventory.yaml'
        ]
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)
        mock_playbook.assert_called_once_with(
            workdir=mock.ANY,
            playbook='prepare-undercloud-backup.yaml',
            inventory=parsed_args.inventory,
            skip_tags='bar_create_recover_image',
            playbook_dir=constants.ANSIBLE_TRIPLEO_PLAYBOOKS,
            verbosity=3,
            extra_vars=None
        )

    @mock.patch('tripleoclient.utils.run_ansible_playbook',
                autospec=True)
    def test_undercloud_backup_inventory(self, mock_playbook):
        arglist = [
            '--inventory',
            '/tmp/test_inventory.yaml'
        ]
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)
        mock_playbook.assert_called_once_with(
            workdir=mock.ANY,
            playbook='cli-undercloud-backup.yaml',
            inventory=parsed_args.inventory,
            skip_tags=None,
            playbook_dir=constants.ANSIBLE_TRIPLEO_PLAYBOOKS,
            verbosity=3,
            extra_vars=None
        )



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import logging
import os

from osc_lib.i18n import _

from tripleoclient import command
from tripleoclient import constants
from tripleoclient import utils


class DownloadConfig(command.Command):
    """Download Overcloud Config"""

    log = logging.getLogger(__name__ + ".DownloadConfig")

    def get_parser(self, prog_name):
        parser = super(DownloadConfig, self).get_parser(prog_name)
        parser.add_argument(
            '--name',
            dest='name',
            default='overcloud',
            help=_('The name of the plan, which is used for the object '
                   'storage container, workflow environment and orchestration '
                   'stack names.'),
        )
        parser.add_argument(
            '--config-dir',
            dest='config_dir',
            default=os.path.join(
                constants.CLOUD_HOME_DIR,
                'tripleo-config'
            ),
            help=_('The directory where the configuration files will be '
                   'pushed'),
        )
        parser.add_argument(
            '--config-type',
            dest='config_type',
            type=list,
            default=None,
            help=_('Type of object config to be extract from the deployment, '
                   'defaults to all keys available'),
        )
        parser.add_argument(
            '--no-preserve-config',
            dest='preserve_config_dir',
            action='store_false',
            default=True,
            help=('If specified, will delete and recreate the --config-dir '
                  'if it already exists. Default is to use the existing dir '
                  'location and overwrite files. Files in --config-dir not '
                  'from the stack will be preserved by default.')
        )
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)" % parsed_args)

        name = parsed_args.name
        config_dir = os.path.abspath(parsed_args.config_dir)
        config_type = parsed_args.config_type
        preserve_config_dir = parsed_args.preserve_config_dir
        extra_vars = {'plan': name,
                      'config_dir': config_dir,
                      'preserve_config': preserve_config_dir}
        if config_type:
            extra_vars['config_type'] = config_type

        with utils.TempDirs() as tmp:
            utils.run_ansible_playbook(
                playbook='cli-config-download-export.yaml',
                inventory='localhost,',
                workdir=tmp,
                playbook_dir=constants.ANSIBLE_TRIPLEO_PLAYBOOKS,
                verbosity=utils.playbook_verbosity(self=self),
                extra_vars=extra_vars)

        print("The TripleO configuration has been successfully generated "
              "into: {0}".format(config_dir))



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
import mock

from osc_lib.tests import utils

from tripleoclient import constants
from tripleoclient import exceptions
from tripleoclient import plugin
from tripleoclient.tests import fakes
from tripleoclient.v1 import overcloud_plan


class TestStringCapture(object):
    def __init__(self):
        self.capture_string = ''

    def write(self, msg):
        self.capture_string = self.capture_string + msg

    def getvalue(self):
        return self.capture_string

    def flush(self):
        return


class TestOvercloudPlanList(utils.TestCommand):

    def setUp(self):
        super(TestOvercloudPlanList, self).setUp()
        self.app.client_manager.tripleoclient = plugin.ClientWrapper(
            instance=fakes.FakeInstanceData
        )
        self.cmd = overcloud_plan.ListPlans(self.app, None)

    @mock.patch("tripleoclient.workflows.plan_management."
                "list_deployment_plans",
                autospec=True)
    def test_list_empty(self, mock_list_plans):
        mock_list_plans.return_value = []

        result = self.cmd.take_action(None)

        self.assertEqual(0, len(result[1]))

    @mock.patch("tripleoclient.workflows.plan_management."
                "list_deployment_plans",
                autospec=True)
    def test_list(self, mock_list_plans):
        mock_list_plans.return_value = (['test-plan-1', 'test-plan-2'])

        result = self.cmd.take_action(None)

        self.assertEqual(1, len(result[0]))
        self.assertEqual([('test-plan-1',), ('test-plan-2',)], result[1])


class TestOvercloudDeletePlan(fakes.FakePlaybookExecution):

    def setUp(self):
        super(TestOvercloudDeletePlan, self).setUp()

        self.cmd = overcloud_plan.DeletePlan(self.app, None)

    @mock.patch("tripleo_common.actions.plan.DeletePlanAction.run",
                return_value=None)
    def test_delete_plan(self, mock_run):
        parsed_args = self.check_parser(self.cmd, ['test-plan'],
                                        [('plans', ['test-plan'])])

        self.cmd.take_action(parsed_args)

    @mock.patch("tripleo_common.actions.plan.DeletePlanAction.run",
                return_value=None)
    def test_delete_multiple_plans(self, mock_run):
        argslist = ['test-plan1', 'test-plan2']
        verifylist = [('plans', ['test-plan1', 'test-plan2'])]
        parsed_args = self.check_parser(self.cmd, argslist, verifylist)

        self.cmd.take_action(parsed_args)


class TestOvercloudCreatePlan(utils.TestCommand):

    def setUp(self):
        super(TestOvercloudCreatePlan, self).setUp()

        app_args = mock.Mock()
        app_args.verbose_level = 1
        self.app.options = fakes.FakeOptions()
        self.cmd = overcloud_plan.CreatePlan(self.app, app_args)
        self.app.client_manager.workflow_engine = mock.Mock()
        self.tripleoclient = mock.Mock()
        self.app.client_manager.tripleoclient = self.tripleoclient

        self.swift = self.app.client_manager.tripleoclient.object_store
        self.swift.get_account = mock.MagicMock()
        self.mock_tar = mock.patch(
            'tripleo_common.utils.tarball.create_tarball',
            autospec=True
        )
        self.mock_tar.start()

    def tearDown(self):
        super(TestOvercloudCreatePlan, self).tearDown()
        self.mock_tar.stop()

    @mock.patch("tripleoclient.utils.run_ansible_playbook", autospec=True)
    @mock.patch('os.chdir', autospec=True)
    @mock.patch('tempfile.mkdtemp', autospec=True)
    def test_create_default_plan(self, mock_tmp, mock_cd, mock_run_playbook):

        # Setup
        arglist = ['overcast']
        verifylist = [
            ('name', 'overcast'),
            ('templates', None)
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # Run
        self.cmd.take_action(parsed_args)

        # Verify
        mock_run_playbook.assert_called_with(
            'cli-create-deployment-plan.yaml',
            'undercloud,',
            mock.ANY,
            constants.ANSIBLE_TRIPLEO_PLAYBOOKS,
            extra_vars={
                "container": "overcast",
                "generate_passwords": True,
                "use_default_templates": True,
                "disable_image_params_prepare": False,
            },
            verbosity=3,
        )

    @mock.patch("tripleoclient.utils.run_ansible_playbook", autospec=True)
    @mock.patch('os.chdir', autospec=True)
    @mock.patch('tempfile.mkdtemp', autospec=True)
    def test_create_custom_plan(self, mock_tmp, mock_cd,
                                mock_run_playbook):

        # Setup
        arglist = ['overcast', '--templates', '/fake/path']
        verifylist = [
            ('name', 'overcast'),
            ('templates', '/fake/path')
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # Run
        self.cmd.take_action(parsed_args)

        # Verify
        mock_run_playbook.assert_called_with(
            'cli-create-deployment-plan.yaml',
            'undercloud,',
            mock.ANY,
            constants.ANSIBLE_TRIPLEO_PLAYBOOKS,
            extra_vars={
                "container": "overcast",
                "generate_passwords": True,
                "use_default_templates": False,
                "disable_image_params_prepare": False,
            },
            verbosity=3,
        )
        self.swift.get_account.assert_called_once()

    @mock.patch("tripleoclient.utils.run_ansible_playbook", autospec=True)
    @mock.patch('os.chdir', autospec=True)
    @mock.patch('tempfile.mkdtemp', autospec=True)
    def test_create_custom_plan_plan_environment_file(
            self, mock_tmp, mock_cd, mock_run_playbook):
        # Setup
        arglist = ['overcast', '--templates', '/fake/path',
                   '-p', 'the_plan_environment.yaml']
        verifylist = [
            ('name', 'overcast'),
            ('templates', '/fake/path'),
            ('plan_environment_file', 'the_plan_environment.yaml')
        ]
        self.app.options = fakes.FakeOptions()
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        mock_open = mock.mock_open()
        # Run
        with mock.patch('six.moves.builtins.open', mock_open):
            self.cmd.take_action(parsed_args)

        mock_open.assert_has_calls(
            [mock.call('the_plan_environment.yaml', 'rb')])

        # Verify
        mock_run_playbook.assert_called_with(
            'cli-create-deployment-plan.yaml',
            'undercloud,',
            mock.ANY,
            constants.ANSIBLE_TRIPLEO_PLAYBOOKS,
            extra_vars={
                "container": "overcast",
                "generate_passwords": True,
                "plan_environment": "the_plan_environment.yaml",
                "use_default_templates": False,
                "disable_image_params_prepare": False,
            },
            verbosity=3,
        )
        self.swift.get_account.assert_called_once()

    @mock.patch("tripleoclient.utils.run_ansible_playbook", autospec=True)
    @mock.patch('os.chdir', autospec=True)
    @mock.patch('tempfile.mkdtemp', autospec=True)
    def test_create_default_plan_with_password_gen_disabled(
            self, mock_tmp, mock_cd, mock_run_playbook):

        # Setup
        arglist = ['overcast', '--disable-password-generation',
                   '--disable-container-prepare']
        verifylist = [
            ('name', 'overcast'),
            ('templates', None),
            ('disable_password_generation', True),
            ('disable_container_prepare', True)
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # Run
        self.app.options = fakes.FakeOptions()
        self.cmd.take_action(parsed_args)
        # Verify
        mock_run_playbook.assert_called_with(
            'cli-create-deployment-plan.yaml',
            'undercloud,',
            mock.ANY,
            constants.ANSIBLE_TRIPLEO_PLAYBOOKS,
            extra_vars={
                "container": "overcast",
                "generate_passwords": False,
                "use_default_templates": True,
                "disable_image_params_prepare": True,
            },
            verbosity=3,
        )

    @mock.patch("tripleoclient.utils.run_ansible_playbook", autospec=True)
    @mock.patch('os.chdir', autospec=True)
    @mock.patch('tempfile.mkdtemp', autospec=True)
    def test_create_custom_plan_from_source_url(
            self, mock_tmp, mock_cd, mock_run_playbook):

        # Setup
        arglist = ['overcast', '--source-url', 'http://tripleo.org/templates']
        verifylist = [
            ('name', 'overcast'),
            ('templates', None),
            ('source_url', 'http://tripleo.org/templates')
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # Run
        self.cmd.take_action(parsed_args)

        # Verify
        mock_run_playbook.assert_called_with(
            'cli-create-deployment-plan.yaml',
            'undercloud,',
            mock.ANY,
            constants.ANSIBLE_TRIPLEO_PLAYBOOKS,
            extra_vars={
                "container": "overcast",
                "generate_passwords": True,
                "use_default_templates": False,
                "source_url": "http://tripleo.org/templates",
                "disable_image_params_prepare": False,
            },
            verbosity=3,
        )


class TestOvercloudDeployPlan(utils.TestCommand):

    def setUp(self):
        super(TestOvercloudDeployPlan, self).setUp()

        app_args = mock.Mock()
        app_args.verbose_level = 1
        self.app.options = fakes.FakeOptions()
        self.cmd = overcloud_plan.DeployPlan(self.app, app_args)

        sleep_patch = mock.patch('time.sleep')
        self.addCleanup(sleep_patch.stop)
        sleep_patch.start()

    @mock.patch("tripleoclient.utils.update_deployment_status", autospec=True)
    @mock.patch("tripleoclient.utils.run_ansible_playbook", autospec=True)
    @mock.patch('tripleoclient.utils.wait_for_stack_ready', autospec=True)
    @mock.patch('os.chdir', autospec=True)
    @mock.patch('tempfile.mkdtemp', autospec=True)
    def test_overcloud_deploy_plan(self, mock_tmp, mock_cd,
                                   mock_for_stack_ready,
                                   mock_run_playbook,
                                   mock_update_status):

        # Setup
        arglist = ['--run-validations', 'overcast']
        verifylist = [
            ('name', 'overcast'),
            ('run_validations', True),
            ('timeout', 240)
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.orch = self.app.client_manager.orchestration = mock.Mock()
        # No existing stack, this is a new deploy.
        self.orch.stacks.get.return_value = None

        mock_for_stack_ready.return_value = True

        # Run
        self.cmd.take_action(parsed_args)

        mock_run_playbook.assert_called_with(
            'cli-deploy-deployment-plan.yaml',
            'undercloud,',
            mock.ANY,
            constants.ANSIBLE_TRIPLEO_PLAYBOOKS,
            timeout=240,
            extra_vars={
                "container": "overcast",
                "run_validations": True,
                "skip_deploy_identifier": False,
            },
            verbosity=3,
        )
        mock_update_status.assert_called()


class TestOvercloudExportPlan(utils.TestCommand):

    def setUp(self):
        super(TestOvercloudExportPlan, self).setUp()
        self.cmd = overcloud_plan.ExportPlan(self.app, None)
        self.app.client_manager = mock.Mock()
        self.clients = self.app.client_manager

        # Mock urlopen
        f = mock.Mock()
        f.read.return_value = 'tarball contents'
        urlopen_patcher = mock.patch('six.moves.urllib.request.urlopen',
                                     return_value=f)
        self.mock_urlopen = urlopen_patcher.start()
        self.addCleanup(self.mock_urlopen.stop)

    @mock.patch(
        'tripleoclient.workflows.plan_management.export_deployment_plan',
        autospec=True)
    def test_export_plan(self, export_deployment_plan_mock):
        parsed_args = self.check_parser(self.cmd, ['test-plan'],
                                        [('plan', 'test-plan')])

        export_deployment_plan_mock.return_value = 'http://fake-url.com'

        with mock.patch('six.moves.builtins.open', mock.mock_open()):
            self.cmd.take_action(parsed_args)

        export_deployment_plan_mock.assert_called_once_with(
            self.clients, 'test-plan')

    @mock.patch('os.path.exists')
    def test_export_plan_outfile_exists(self, exists_mock):
        parsed_args = self.check_parser(self.cmd, ['test-plan'],
                                        [('plan', 'test-plan')])

        exists_mock.return_value = True

        self.assertRaises(exceptions.PlanExportError,
                          self.cmd.take_action, parsed_args)

    @mock.patch(
        'tripleoclient.workflows.plan_management.export_deployment_plan',
        autospec=True)
    @mock.patch('os.path.exists')
    def test_export_plan_outfile_exists_with_overwrite(
            self, exists_mock, export_deployment_plan_mock):
        arglist = ['-f', 'test-plan']
        verifylist = [
            ('plan', 'test-plan'),
            ('force_overwrite', True)
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exists_mock.return_value = True
        export_deployment_plan_mock.return_value = 'http://fake-url.com'

        with mock.patch('six.moves.builtins.open', mock.mock_open()):
            self.cmd.take_action(parsed_args)

        export_deployment_plan_mock.assert_called_once_with(
            self.clients, 'test-plan')



#   Copyright 2019 Red Hat, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import copy

from osc_lib.i18n import _
from oslo_config import cfg

from tripleoclient import constants

from tripleoclient.config.standalone import StandaloneConfig

CONF = cfg.CONF


class MinionConfig(StandaloneConfig):
    def get_minion_service_opts(self, heat_engine=True,
                                ironic_conductor=False):
        _opts = [
            cfg.BoolOpt('enable_heat_engine',
                        default=heat_engine,
                        help=_(
                            'Whether to install the Heat Engine service.')),
            cfg.BoolOpt('enable_ironic_conductor',
                        default=ironic_conductor,
                        help=_(
                            'Whether to install the Ironic Conductor service. '
                            'This is currently disabled by default.')),
        ]
        return self.sort_opts(_opts)

    def get_base_opts(self):
        _base_opts = super(MinionConfig, self).get_base_opts()
        _opts = [
            cfg.StrOpt('minion_log_file',
                       default=constants.MINION_LOG_FILE,
                       help=_(
                           'The path to a log file to store the '
                           'install/upgrade logs.'),
                       ),
            cfg.StrOpt('minion_hostname',
                       help=_(
                           'Fully qualified hostname (including domain) to '
                           'set on the Undercloud Minion. If left unset, the '
                           'current hostname will be used, but the user is '
                           'responsible for configuring all system hostname '
                           'settings appropriately.  If set, the Undercloud '
                           'Minion install will configure all system hostname '
                           'settings.'),
                       ),
            cfg.StrOpt('minion_local_ip',
                       default='192.168.24.50/24',
                       help=_(
                           'IP information for the interface on the '
                           'Undercloud Minion.  The IP portion '
                           'of the value will be assigned to the network '
                           'interface defined by local_interface, with the '
                           'netmask defined by the prefix portion of the '
                           'value.')
                       ),
            cfg.ListOpt('minion_nameservers',
                        default=[],
                        help=_(
                            'DNS nameserver(s) to configure on the Undercloud '
                            'Minion.')
                        ),
            cfg.ListOpt('minion_ntp_servers',
                        default=['0.pool.ntp.org', '1.pool.ntp.org',
                                 '2.pool.ntp.org', '3.pool.ntp.org'],
                        help=_('List of ntp servers to use.')),
            cfg.StrOpt('minion_timezone', default=None,
                       help=_('Host timezone to be used. If no timezone is '
                              'specified, the existing timezone configuration '
                              'is used.')),
            cfg.StrOpt('minion_service_certificate',
                       default='',
                       help=_(
                           'TODO: '
                           'Certificate file to use for OpenStack service SSL '
                           'connections.  Setting this enables SSL for the '
                           'OpenStack API endpoints, leaving it unset '
                           'disables SSL.')
                       ),
            cfg.StrOpt('minion_password_file',
                       default='tripleo-undercloud-passwords.yaml',
                       help=_(
                           'The name of the file to look for the passwords '
                           'used to connect to the Undercloud. We assume '
                           'this file is in the folder where the command '
                           'is executed if a fully qualified path is not '
                           'provided.')
                       ),
            cfg.StrOpt('minion_undercloud_output_file',
                       default='tripleo-undercloud-outputs.yaml',
                       help=_(
                           'The name of the file to look for the Undercloud '
                           'output file that contains configuration '
                           'information. We assume this file is in the folder '
                           'where the command is executed if a fully '
                           'qualified path is not provided.')
                       ),

            cfg.StrOpt('minion_local_interface',
                       default='eth1',
                       help=_('Network interface on the Undercloud Minion '
                              'that will be used for the services.')
                       ),
            cfg.IntOpt('minion_local_mtu',
                       default=1500,
                       help=_('MTU to use for the local_interface.')
                       ),
            cfg.BoolOpt('minion_debug',
                        default=True,
                        help=_(
                            'Whether to enable the debug log level for '
                            'OpenStack services and Container Image Prepare '
                            'step.')
                        ),
            cfg.BoolOpt('minion_enable_selinux',
                        default=True,
                        help=_('Enable or disable SELinux during the '
                               'deployment.')),
            cfg.BoolOpt('minion_enable_validations',
                        default=True,
                        help=_(
                            'Run pre-flight checks when installing or '
                            'upgrading.')
                        ),

        ]
        return self.sort_opts(_base_opts + _opts)

    def get_opts(self):
        _base_opts = self.get_base_opts()
        _service_opts = self.get_minion_service_opts()
        return self.sort_opts(_base_opts + _service_opts)


def list_opts():
    """List config opts for oslo config generator"""
    return [(None, copy.deepcopy(MinionConfig().get_opts()))]


def load_global_config():
    """Register MinionConfig options into global config"""
    _opts = MinionConfig().get_opts()
    CONF.register_opts(_opts)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging

import yaml

from tripleo_common.actions import plan
# TODO(cloudnull): Convert to a swiftutils in tripleo-common
# from tripleo_common.utils import swift as swiftutils


LOG = logging.getLogger(__name__)


def list_available_roles(clients, container='overcloud'):
    """Return a list of available roles.

    :param clients: openstack clients
    :type clients: Object

    :param container: Name of swift object container
    :type container: String

    :returns: List
    """

    LOG.info('Pulling role list from: {}'.format(container))
    obj_client = clients.tripleoclient.object_store
    available_yaml_roles = list()
    LOG.info('Indexing roles from: {}'.format(container))
    # TODO(cloudnull): Convert to a swiftutils in tripleo-common
    for obj in obj_client.get_container(container)[-1]:
        name = obj['name']
        if name.startswith('roles/') and name.endswith(('yml', 'yaml')):
            role_data = yaml.safe_load(
                obj_client.get_object(container, name)[-1]
            )
            available_yaml_roles.append(role_data[0])

    return available_yaml_roles


def list_roles(clients, container, detail=False):
    """Return a list of roles.

    :param clients: openstack clients
    :type clients: Object

    :param container: Name of swift object container
    :type container: String

    :param detail: Enable or disable extra detail
    :type detail: Boolean

    :returns: List
    """

    context = clients.tripleoclient.create_mistral_context()
    LOG.info('Pulling roles from: {}'.format(container))
    return plan.ListRolesAction(
        container=container,
        detail=detail
    ).run(context=context)



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import mock

from osc_lib.tests import utils

from tripleoclient.tests.fakes import FakeStackObject
from tripleoclient.workflows import deployment


class TestDeploymentWorkflows(utils.TestCommand):

    def setUp(self):
        super(TestDeploymentWorkflows, self).setUp()
        self.tripleoclient = mock.Mock()

    @mock.patch('shutil.rmtree')
    @mock.patch('os.chdir')
    @mock.patch('tripleoclient.utils.tempfile')
    @mock.patch('tripleoclient.utils.run_ansible_playbook',
                autospec=True)
    def test_enable_ssh_admin(self, mock_rmtree, mock_chdir, mock_tempfile,
                              mock_playbook):
        hosts = 'a', 'b', 'c'
        ssh_user = 'test-user'
        ssh_key = 'test-key'
        timeout = 30

        deployment.enable_ssh_admin(
            FakeStackObject,
            hosts,
            ssh_user,
            ssh_key,
            timeout
        )

        # once for ssh-keygen, then twice per host
        self.assertEqual(1, mock_playbook.call_count)

    @mock.patch('tripleoclient.utils.get_blacklisted_ip_addresses')
    @mock.patch('tripleoclient.utils.get_role_net_ip_map')
    def test_get_overcloud_hosts(self, mock_role_net_ip_map,
                                 mock_blacklisted_ip_addresses):
        stack = mock.Mock()
        mock_role_net_ip_map.return_value = {
            'Controller': {
                'ctlplane': ['1.1.1.1', '2.2.2.2', '3.3.3.3'],
                'external': ['4.4.4.4', '5.5.5.5', '6.6.6.6']},
            'Compute': {
                'ctlplane': ['7.7.7.7', '8.8.8.8', '9.9.9.9'],
                'external': ['10.10.10.10', '11.11.11.11', '12.12.12.12']},
        }
        mock_blacklisted_ip_addresses.return_value = []

        ips = deployment.get_overcloud_hosts(stack, 'ctlplane')
        expected = ['1.1.1.1', '2.2.2.2', '3.3.3.3',
                    '7.7.7.7', '8.8.8.8', '9.9.9.9']
        self.assertEqual(sorted(expected), sorted(ips))

        ips = deployment.get_overcloud_hosts(stack, 'external')
        expected = ['4.4.4.4', '5.5.5.5', '6.6.6.6',
                    '10.10.10.10', '11.11.11.11', '12.12.12.12']
        self.assertEqual(sorted(expected), sorted(ips))

    @mock.patch('tripleoclient.utils.get_blacklisted_ip_addresses')
    @mock.patch('tripleoclient.utils.get_role_net_ip_map')
    def test_get_overcloud_hosts_with_blacklist(
            self, mock_role_net_ip_map,
            mock_blacklisted_ip_addresses):
        stack = mock.Mock()
        stack.output_show.return_value = []
        mock_role_net_ip_map.return_value = {
            'Controller': {
                'ctlplane': ['1.1.1.1', '2.2.2.2', '3.3.3.3'],
                'external': ['4.4.4.4', '5.5.5.5', '6.6.6.6']},
            'Compute': {
                'ctlplane': ['7.7.7.7', '8.8.8.8', '9.9.9.9'],
                'external': ['10.10.10.10', '11.11.11.11', '12.12.12.12']},
        }

        mock_blacklisted_ip_addresses.return_value = ['8.8.8.8']
        ips = deployment.get_overcloud_hosts(stack, 'ctlplane')
        expected = ['1.1.1.1', '2.2.2.2', '3.3.3.3',
                    '7.7.7.7', '9.9.9.9']
        self.assertEqual(sorted(expected), sorted(ips))

        ips = deployment.get_overcloud_hosts(stack, 'external')
        expected = ['4.4.4.4', '5.5.5.5', '6.6.6.6',
                    '10.10.10.10', '12.12.12.12']
        self.assertEqual(sorted(expected), sorted(ips))

        mock_blacklisted_ip_addresses.return_value = ['7.7.7.7', '9.9.9.9',
                                                      '2.2.2.2']
        ips = deployment.get_overcloud_hosts(stack, 'external')
        expected = ['4.4.4.4', '6.6.6.6', '11.11.11.11']
        self.assertEqual(sorted(expected), sorted(ips))

    @mock.patch('tripleoclient.utils.run_ansible_playbook',
                autospec=True)
    def test_config_download_already_in_progress_for_diff_stack(
            self, mock_playbook):
        log = mock.Mock()
        stack = mock.Mock()
        stack.stack_name = 'stacktest'
        stack.output_show.return_value = {'output': {'output_value': []}}
        clients = mock.Mock()
        deployment.config_download(
            log, clients, stack, 'templates', 'ssh_user',
            'ssh_key', 'ssh_networks', 'output_dir', False,
            'timeout')

        self.assertEqual(3, mock_playbook.call_count)



#   Copyright 2015 Red Hat, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#


import ansible_runner
import argparse
import datetime
import logging
import mock
import os
import os.path
import shutil
import socket
import subprocess
import tempfile

import sys

from heatclient import exc as hc_exc

from uuid import uuid4

from testscenarios import TestWithScenarios
from unittest import TestCase
import yaml

from tripleoclient import exceptions
from tripleoclient import utils

from tripleoclient.tests import base
from tripleoclient.tests import fakes

from six.moves.configparser import ConfigParser
from six.moves.urllib import error as url_error

from ansible_runner import Runner


class TestRunAnsiblePlaybook(TestCase):
    def setUp(self):
        self.unlink_patch = mock.patch('os.unlink')
        self.addCleanup(self.unlink_patch.stop)
        self.unlink_patch.start()
        self.mock_log = mock.Mock('logging.getLogger')
        self.ansible_playbook_cmd = "ansible-playbook"
        self.orig_workdir = utils.constants.DEFAULT_WORK_DIR
        utils.constants.DEFAULT_WORK_DIR = utils.TempDirs().dir
        utils.makedirs(
            os.path.join(
                utils.constants.DEFAULT_WORK_DIR,
                'overcloud'
            )
        )

    def tearDown(self):
        utils.constants.DEFAULT_WORK_DIR = self.orig_workdir

    @mock.patch('os.path.exists', return_value=False)
    @mock.patch('tripleoclient.utils.run_command_and_log')
    @mock.patch('ansible_runner.utils.dump_artifact', autospec=True,
                return_value="/foo/inventory.yaml")
    def test_no_playbook(self, mock_dump_artifact, mock_run, mock_exists):
        self.assertRaises(
            RuntimeError,
            utils.run_ansible_playbook,
            'non-existing.yaml',
            'localhost,',
            '/tmp'
        )
        mock_exists.assert_called_with('/tmp/non-existing.yaml')
        mock_run.assert_not_called()

    @mock.patch('tempfile.mkstemp', return_value=('foo', '/tmp/fooBar.cfg'))
    @mock.patch('os.path.exists', return_value=True)
    @mock.patch('os.makedirs')
    @mock.patch.object(
        Runner,
        'run',
        return_value=fakes.fake_ansible_runner_run_return(rc=1)
    )
    @mock.patch('ansible_runner.utils.dump_artifact', autospec=True,
                return_value="/foo/inventory.yaml")
    def test_subprocess_error(self, mock_dump_artifact,
                              mock_run, mock_mkdirs, mock_exists,
                              mock_mkstemp):
        with self.assertRaises(RuntimeError):
            utils.run_ansible_playbook(
                'existing.yaml',
                'localhost,',
                '/tmp'
            )

    @mock.patch('tempfile.mkstemp', return_value=('foo', '/tmp/fooBar.cfg'))
    @mock.patch('os.path.exists', return_value=True)
    @mock.patch('os.makedirs')
    @mock.patch.object(
        Runner,
        'run',
        return_value=fakes.fake_ansible_runner_run_return()
    )
    @mock.patch('ansible_runner.utils.dump_artifact', autospec=True,
                return_value="/foo/inventory.yaml")
    def test_run_success_default(self, mock_dump_artifact, mock_run,
                                 mock_mkdirs, mock_exists, mock_mkstemp):
        utils.run_ansible_playbook(
            playbook='existing.yaml',
            inventory='localhost,',
            workdir='/tmp'
        )

    @mock.patch('os.path.exists', return_value=True)
    @mock.patch('os.makedirs')
    @mock.patch.object(
        Runner,
        'run',
        return_value=fakes.fake_ansible_runner_run_return()
    )
    @mock.patch('ansible_runner.utils.dump_artifact', autospec=True,
                return_value="/foo/inventory.yaml")
    def test_run_success_ansible_cfg(self, mock_dump_artifact, mock_run,
                                     mock_mkdirs, mock_exists):
        utils.run_ansible_playbook(
            playbook='existing.yaml',
            inventory='localhost,',
            workdir='/tmp'
        )

    @mock.patch('tempfile.mkstemp', return_value=('foo', '/tmp/fooBar.cfg'))
    @mock.patch('os.path.exists', return_value=True)
    @mock.patch('os.makedirs')
    @mock.patch.object(
        Runner,
        'run',
        return_value=fakes.fake_ansible_runner_run_return()
    )
    @mock.patch('ansible_runner.utils.dump_artifact', autospec=True,
                return_value="/foo/inventory.yaml")
    def test_run_success_connection_local(self, mock_dump_artifact, mock_run,
                                          mock_mkdirs, mock_exists,
                                          mock_mkstemp):
        utils.run_ansible_playbook(
            playbook='existing.yaml',
            inventory='localhost,',
            workdir='/tmp',
            connection='local'
        )

    @mock.patch('os.makedirs', return_value=None)
    @mock.patch('tempfile.mkstemp', return_value=('foo', '/tmp/fooBar.cfg'))
    @mock.patch('os.path.exists', return_value=True)
    @mock.patch.object(
        Runner,
        'run',
        return_value=fakes.fake_ansible_runner_run_return()
    )
    @mock.patch('ansible_runner.utils.dump_artifact', autospec=True,
                return_value="/foo/inventory.yaml")
    def test_run_success_gathering_policy(self, mock_dump_artifact, mock_run,
                                          mock_exists, mock_mkstemp,
                                          mock_makedirs):
        utils.run_ansible_playbook(
            playbook='existing.yaml',
            inventory='localhost,',
            workdir='/tmp',
            connection='local',
            gathering_policy='smart'
        )

    @mock.patch('os.makedirs', return_value=None)
    @mock.patch('tempfile.mkstemp', return_value=('foo', '/tmp/fooBar.cfg'))
    @mock.patch('os.path.exists', return_value=True)
    @mock.patch.object(
        Runner,
        'run',
        return_value=fakes.fake_ansible_runner_run_return()
    )
    @mock.patch('ansible_runner.utils.dump_artifact', autospec=True,
                return_value="/foo/inventory.yaml")
    def test_run_success_extra_vars(self, mock_dump_artifact, mock_run,
                                    mock_exists, mock_mkstemp, mock_makedirs):
        arglist = {
            'var_one': 'val_one',
        }
        utils.run_ansible_playbook(
            playbook='existing.yaml',
            inventory='localhost,',
            workdir='/tmp',
            connection='local',
            gathering_policy='smart',
            extra_vars=arglist
        )

    @mock.patch('six.moves.builtins.open')
    @mock.patch('tripleoclient.utils.makedirs')
    @mock.patch('os.path.exists', side_effect=(False, True, True))
    def test_run_with_timeout(self, mock_exists, mock_mkdir, mock_open):
        ansible_runner.ArtifactLoader = mock.MagicMock()
        ansible_runner.Runner.run = mock.MagicMock(return_value=('', 0))
        ansible_runner.runner_config = mock.MagicMock()
        utils.run_ansible_playbook(
            playbook='existing.yaml',
            inventory='localhost,',
            workdir='/tmp',
            timeout=42
        )
        self.assertIn(mock.call('/tmp/env/settings', 'w'),
                      mock_open.mock_calls)
        self.assertIn(
            mock.call().__enter__().write('job_timeout: 2520\n'),  # 42m * 60
            mock_open.mock_calls)

    @mock.patch('six.moves.builtins.open')
    @mock.patch('tripleoclient.utils.makedirs')
    @mock.patch('os.path.exists', side_effect=(False, True, True))
    def test_run_with_extravar_file(self, mock_exists, mock_mkdir, mock_open):
        ansible_runner.ArtifactLoader = mock.MagicMock()
        ansible_runner.Runner.run = mock.MagicMock(return_value=('', 0))
        ansible_runner.runner_config = mock.MagicMock()
        utils.run_ansible_playbook(
            playbook='existing.yaml',
            inventory='localhost,',
            workdir='/tmp',
            extra_vars_file={
                'foo': 'bar',
                'things': {
                    'more': 'options'
                },
                'num': 42
            }
        )
        self.assertIn(
            mock.call('/tmp/env/extravars', 'w'),
            mock_open.mock_calls
        )
        self.assertIn(
            mock.call().__enter__().write(
                'foo: bar\nnum: 42\nthings:\n  more: options\n'
            ),
            mock_open.mock_calls
        )


class TestRunCommandAndLog(TestCase):
    def setUp(self):
        self.mock_logger = mock.Mock(spec=logging.Logger)

        self.mock_process = mock.Mock()
        self.mock_process.stdout.readline.side_effect = ['foo\n', 'bar\n']
        self.mock_process.wait.side_effect = [0]
        self.mock_process.returncode = 0

        mock_sub = mock.patch('subprocess.Popen',
                              return_value=self.mock_process)
        self.mock_popen = mock_sub.start()
        self.addCleanup(mock_sub.stop)

        self.cmd = ['exit', '0']
        self.e_cmd = ['exit', '1']
        self.log_calls = [mock.call('foo'),
                          mock.call('bar')]

    def test_success_default(self):
        retcode = utils.run_command_and_log(self.mock_logger, self.cmd)
        self.mock_popen.assert_called_once_with(self.cmd,
                                                stdout=subprocess.PIPE,
                                                stderr=subprocess.STDOUT,
                                                shell=False,
                                                cwd=None, env=None)
        self.assertEqual(retcode, 0)
        self.mock_logger.warning.assert_has_calls(self.log_calls,
                                                  any_order=False)

    @mock.patch('subprocess.Popen')
    def test_error_subprocess(self, mock_popen):
        mock_process = mock.Mock()
        mock_process.stdout.readline.side_effect = ['Error\n']
        mock_process.wait.side_effect = [1]
        mock_process.returncode = 1

        mock_popen.return_value = mock_process

        retcode = utils.run_command_and_log(self.mock_logger, self.e_cmd)
        mock_popen.assert_called_once_with(self.e_cmd, stdout=subprocess.PIPE,
                                           stderr=subprocess.STDOUT,
                                           shell=False, cwd=None,
                                           env=None)

        self.assertEqual(retcode, 1)
        self.mock_logger.warning.assert_called_once_with('Error')

    def test_success_env(self):
        test_env = os.environ.copy()
        retcode = utils.run_command_and_log(self.mock_logger, self.cmd,
                                            env=test_env)
        self.mock_popen.assert_called_once_with(self.cmd,
                                                stdout=subprocess.PIPE,
                                                stderr=subprocess.STDOUT,
                                                shell=False,
                                                cwd=None, env=test_env)
        self.assertEqual(retcode, 0)
        self.mock_logger.warning.assert_has_calls(self.log_calls,
                                                  any_order=False)

    def test_success_cwd(self):
        test_cwd = '/usr/local/bin'
        retcode = utils.run_command_and_log(self.mock_logger, self.cmd,
                                            cwd=test_cwd)
        self.mock_popen.assert_called_once_with(self.cmd,
                                                stdout=subprocess.PIPE,
                                                stderr=subprocess.STDOUT,
                                                shell=False,
                                                cwd=test_cwd, env=None)
        self.assertEqual(retcode, 0)
        self.mock_logger.warning.assert_has_calls(self.log_calls,
                                                  any_order=False)

    def test_success_no_retcode(self):
        run = utils.run_command_and_log(self.mock_logger, self.cmd,
                                        retcode_only=False)
        self.mock_popen.assert_called_once_with(self.cmd,
                                                stdout=subprocess.PIPE,
                                                stderr=subprocess.STDOUT,
                                                shell=False,
                                                cwd=None, env=None)
        self.assertEqual(run, self.mock_process)
        self.mock_logger.warning.assert_not_called()


class TestWaitForStackUtil(TestCase):
    def setUp(self):
        self.mock_orchestration = mock.Mock()
        sleep_patch = mock.patch('time.sleep')
        self.addCleanup(sleep_patch.stop)
        sleep_patch.start()

    def mock_event(self, resource_name, id, resource_status_reason,
                   resource_status, event_time):
        e = mock.Mock()
        e.resource_name = resource_name
        e.id = id
        e.resource_status_reason = resource_status_reason
        e.resource_status = resource_status
        e.event_time = event_time
        return e

    @mock.patch("heatclient.common.event_utils.get_events")
    def test_wait_for_stack_ready(self, mock_el):
        stack = mock.Mock()
        stack.stack_name = 'stack'
        stack.stack_status = "CREATE_COMPLETE"
        self.mock_orchestration.stacks.get.return_value = stack

        complete = utils.wait_for_stack_ready(self.mock_orchestration, 'stack')
        self.assertTrue(complete)

    @mock.patch("time.sleep")
    @mock.patch("heatclient.common.event_utils.poll_for_events")
    @mock.patch("tripleoclient.utils.get_stack")
    def test_wait_for_stack_ready_retry(self, mock_get_stack, mock_poll,
                                        mock_time):
        stack = mock.Mock()
        stack.stack_name = 'stack'
        stack.stack_id = 'id'
        stack.stack_status = "CREATE_COMPLETE"
        mock_get_stack.return_value = stack
        mock_poll.side_effect = [hc_exc.HTTPException(code=504),
                                 ("CREATE_COMPLETE", "ready retry message")]

        complete = utils.wait_for_stack_ready(self.mock_orchestration, 'stack')
        self.assertTrue(complete)

    @mock.patch("time.sleep")
    @mock.patch("heatclient.common.event_utils.poll_for_events")
    @mock.patch("tripleoclient.utils.get_stack")
    def test_wait_for_stack_ready_retry_fail(self, mock_get_stack, mock_poll,
                                             mock_time):
        stack = mock.Mock()
        stack.stack_name = 'stack'
        stack.stack_id = 'id'
        stack.stack_status = "CREATE_COMPLETE"
        mock_get_stack.return_value = stack
        mock_poll.side_effect = hc_exc.HTTPException(code=504)

        self.assertRaises(RuntimeError,
                          utils.wait_for_stack_ready,
                          self.mock_orchestration, 'stack')

    @mock.patch("time.sleep")
    @mock.patch("heatclient.common.event_utils.poll_for_events")
    @mock.patch("tripleoclient.utils.get_stack")
    def test_wait_for_stack_ready_server_fail(self, mock_get_stack, mock_poll,
                                              mock_time):
        stack = mock.Mock()
        stack.stack_name = 'stack'
        stack.stack_id = 'id'
        stack.stack_status = "CREATE_COMPLETE"
        mock_get_stack.return_value = stack
        mock_poll.side_effect = hc_exc.HTTPException(code=500)

        self.assertRaises(RuntimeError,
                          utils.wait_for_stack_ready,
                          self.mock_orchestration, 'stack')

    def test_wait_for_stack_ready_no_stack(self):
        self.mock_orchestration.stacks.get.return_value = None

        complete = utils.wait_for_stack_ready(self.mock_orchestration, 'stack')

        self.assertFalse(complete)

    @mock.patch("heatclient.common.event_utils.get_events")
    def test_wait_for_stack_ready_failed(self, mock_el):
        stack = mock.Mock()
        stack.stack_name = 'stack'
        stack.stack_status = "CREATE_FAILED"
        self.mock_orchestration.stacks.get.return_value = stack

        complete = utils.wait_for_stack_ready(self.mock_orchestration, 'stack')

        self.assertFalse(complete)

    @mock.patch("heatclient.common.event_utils.poll_for_events")
    def test_wait_for_stack_in_progress(self, mock_poll_for_events):

        mock_poll_for_events.return_value = ("CREATE_IN_PROGRESS", "MESSAGE")

        stack = mock.Mock()
        stack.stack_name = 'stack'
        stack.stack_status = 'CREATE_IN_PROGRESS'
        self.mock_orchestration.stacks.get.return_value = stack

        result = utils.wait_for_stack_ready(self.mock_orchestration, 'stack')
        self.assertEqual(False, result)

    def test_check_heat_network_config(self):
        stack_reg = {
            'OS::TripleO::Controller::Net::SoftwareConfig': 'val',
            'OS::TripleO::Compute::Net::SoftwareConfig': 'val',
        }
        env_reg = {
            'OS::TripleO::Controller::Net::SoftwareConfig': 'val',
            'OS::TripleO::Compute::Net::SoftwareConfig': 'val',
        }
        mock_stack = mock.MagicMock()
        mock_stack.environment = mock.MagicMock()
        mock_stack.environment.return_value = {
            'resource_registry':  stack_reg,
        }
        env = {
            'resource_registry':  env_reg
        }

        self.assertRaises(exceptions.InvalidConfiguration,
                          utils.check_nic_config_with_ansible,
                          mock_stack, env)

    def test_check_heat_missing_network_config(self):
        stack_reg = {
            'OS::TripleO::Controller::Net::SoftwareConfig': 'val',
            'OS::TripleO::Compute::Net::SoftwareConfig': 'val',
        }
        env_reg = {
            'OS::TripleO::Controller::Net::SoftwareConfig': 'OS::Heat::None',
        }
        mock_stack = mock.MagicMock()
        mock_stack.environment = mock.MagicMock()
        mock_stack.environment.return_value = {
            'resource_registry':  stack_reg,
        }
        env = {
            'resource_registry':  env_reg
        }

        self.assertRaises(exceptions.InvalidConfiguration,
                          utils.check_nic_config_with_ansible,
                          mock_stack, env)

    def test_check_heat_none_network_config(self):
        stack_reg = {
            'OS::TripleO::Controller::Net::SoftwareConfig': 'val',
            'OS::TripleO::Compute::Net::SoftwareConfig': 'val',
        }
        env_reg = {
            'OS::TripleO::Controller::Net::SoftwareConfig': 'OS::Heat::None',
            'OS::TripleO::Compute::Net::SoftwareConfig': 'OS::Heat::None',
        }
        mock_stack = mock.MagicMock()
        mock_stack.environment = mock.MagicMock()
        mock_stack.environment.return_value = {
            'resource_registry':  stack_reg,
        }
        env = {
            'resource_registry':  env_reg,
            'parameter_defaults': {'NetworkConfigWithAnsible': True}
        }
        utils.check_nic_config_with_ansible(mock_stack, env)

    def test_check_heat_network_config_no_ansible(self):
        stack_reg = {
            'OS::TripleO::Controller::Net::SoftwareConfig': 'val',
            'OS::TripleO::Compute::Net::SoftwareConfig': 'val',
        }
        env_reg = {
            'OS::TripleO::Controller::Net::SoftwareConfig': 'val',
            'OS::TripleO::Compute::Net::SoftwareConfig': 'val',
        }
        mock_stack = mock.MagicMock()
        mock_stack.environment = mock.MagicMock()
        mock_stack.environment.return_value = {
            'resource_registry':  stack_reg,
        }
        env = {
            'resource_registry':  env_reg,
            'parameter_defaults': {'NetworkConfigWithAnsible': False}
        }
        utils.check_nic_config_with_ansible(mock_stack, env)

    def test_check_stack_network_matches_env_files(self):
        stack_reg = {
            'OS::TripleO::Network': 'val',
            'OS::TripleO::Network::External': 'val',
            'OS::TripleO::Network::ExtraConfig': 'OS::Heat::None',
            'OS::TripleO::Network::InternalApi': 'val',
            'OS::TripleO::Network::Port::InternalApi': 'val',
            'OS::TripleO::Network::Management': 'val',
            'OS::TripleO::Network::Storage': 'val',
            'OS::TripleO::Network::StorageMgmt': 'val',
            'OS::TripleO::Network::Tenant': 'val'
        }
        env_reg = {
            'OS::TripleO::Network': 'newval',
            'OS::TripleO::Network::External': 'newval',
            'OS::TripleO::Network::ExtraConfig': 'OS::Heat::None',
            'OS::TripleO::Network::InternalApi': 'newval',
            'OS::TripleO::Network::Management': 'newval',
            'OS::TripleO::Network::Storage': 'val',
            'OS::TripleO::Network::StorageMgmt': 'val',
            'OS::TripleO::Network::Tenant': 'val'
        }
        mock_stack = mock.MagicMock()
        mock_stack.environment = mock.MagicMock()
        mock_stack.environment.return_value = {
            'resource_registry':  stack_reg
        }
        env = {
            'resource_registry':  env_reg
        }
        utils.check_stack_network_matches_env_files(mock_stack, env)

    def test_check_stack_network_matches_env_files_fail(self):
        stack_reg = {
            'OS::TripleO::LoggingConfiguration': 'val',
            'OS::TripleO::Network': 'val',
            'OS::TripleO::Network::External': 'val',
            'OS::TripleO::Network::ExtraConfig': 'OS::Heat::None',
            'OS::TripleO::Network::InternalApi': 'val',
            'OS::TripleO::Network::Port::InternalApi': 'val',
            'OS::TripleO::Network::Management': 'val',
            'OS::TripleO::Network::Storage': 'val',
            'OS::TripleO::Network::StorageMgmt': 'val',
            'OS::TripleO::Network::Tenant': 'val'
        }
        env_reg = {
            'OS::TripleO::LoggingConfiguration': 'newval',
            'OS::TripleO::Network': 'newval',
            'OS::TripleO::Network::InternalApi': 'newval'
        }
        mock_stack = mock.MagicMock()
        mock_stack.environment = mock.MagicMock()
        mock_stack.environment.return_value = {
            'resource_registry':  stack_reg
        }
        env = {
            'resource_registry':  env_reg
        }
        with self.assertRaises(exceptions.InvalidConfiguration):
            utils.check_stack_network_matches_env_files(mock_stack, env)

    def test_check_ceph_fsid_matches_env_files(self):
        stack_params = {
            'CephClusterFSID': 'ceph_fsid_val',
            'key1': 'val1',
            'key2': 'val2',
        }
        mock_stack = mock.MagicMock()
        mock_stack.environment = mock.MagicMock()
        mock_stack.environment.return_value = {
            'parameter_defaults': stack_params
        }
        provided_env = {
            'parameter_defaults': {
                'CephClusterFSID': mock_stack.environment()
                                             .get('parameter_defaults', {})
                                             .get('CephClusterFSID', False),
                'key1': 'val1',
                'key2': 'val2',
            }
        }
        utils.check_ceph_fsid_matches_env_files(mock_stack, provided_env)

    def test_check_ceph_fsid_matches_env_files_fail(self):
        stack_params = {
            'CephClusterFSID': 'ceph_fsid_val',
            'key1': 'val1',
            'key2': 'val2',
        }
        provided_env = {
            'parameter_defaults': {
                'CephClusterFSID': 'new_or_wrong_fsid_val',
                'key1': 'val1',
                'key2': 'val2',
            }
        }
        mock_stack = mock.MagicMock()
        mock_stack.environment = mock.MagicMock()
        mock_stack.environment.return_value = {
            'parameter_defaults': stack_params
        }
        with self.assertRaises(exceptions.InvalidConfiguration):
            utils.check_ceph_fsid_matches_env_files(mock_stack, provided_env)

    @mock.patch('subprocess.check_call')
    @mock.patch('os.path.exists')
    def test_remove_known_hosts(self, mock_exists, mock_check_call):

        mock_exists.return_value = True

        utils.remove_known_hosts('192.168.0.1')
        known_hosts = os.path.expanduser("~/.ssh/known_hosts")

        mock_check_call.assert_called_with(
            ['ssh-keygen', '-R', '192.168.0.1', '-f', known_hosts])

    @mock.patch('subprocess.check_call')
    @mock.patch('os.path.exists')
    def test_remove_known_hosts_no_file(self, mock_exists, mock_check_call):

        mock_exists.return_value = False

        utils.remove_known_hosts('192.168.0.1')

        mock_check_call.assert_not_called()

    def test_empty_file_checksum(self):
        # Used a NamedTemporaryFile since it's deleted when the file is closed.
        with tempfile.NamedTemporaryFile() as empty_temp_file:
            self.assertEqual(utils.file_checksum(empty_temp_file.name),
                             'd41d8cd98f00b204e9800998ecf8427e')

    def test_non_empty_file_checksum(self):
        # Used a NamedTemporaryFile since it's deleted when the file is closed.
        with tempfile.NamedTemporaryFile() as temp_file:
            temp_file.write(b'foo')
            temp_file.flush()

            self.assertEqual(utils.file_checksum(temp_file.name),
                             'acbd18db4cc2f85cedef654fccc4a4d8')

    def test_shouldnt_checksum_open_special_files(self):
        self.assertRaises(ValueError, utils.file_checksum, '/dev/random')
        self.assertRaises(ValueError, utils.file_checksum, '/dev/zero')


class TestEnsureRunAsNormalUser(TestCase):

    @mock.patch('os.geteuid')
    def test_ensure_run_as_normal_user(self, os_geteuid_mock):
        os_geteuid_mock.return_value = 1000
        self.assertIsNone(utils.ensure_run_as_normal_user())

    @mock.patch('os.geteuid')
    def test_ensure_run_as_normal_user_root(self, os_geteuid_mock):
        os_geteuid_mock.return_value = 0
        self.assertRaises(exceptions.RootUserExecution,
                          utils.ensure_run_as_normal_user)

    @mock.patch('getpass.getuser')
    def test_get_deployment_user(self, mock_getpass):
        mock_getpass.return_value = 'stack'
        u = utils.get_deployment_user()
        self.assertEqual('stack', u)


class TestCreateTempestDeployerInput(TestCase):

    def test_create_tempest_deployer_input(self):
        with tempfile.NamedTemporaryFile() as cfgfile:
            filepath = cfgfile.name
            utils.create_tempest_deployer_input(filepath)
            with open(filepath, 'rt') as f:
                cfg = f.read()
            # Just make a simple test, to make sure it created a proper file:
            self.assertIn(
                '[volume-feature-enabled]\nbootable = true', cfg)


class TestGetStackOutputItem(TestCase):

    def test_get_stack_output_item(self):
        stack = mock.MagicMock()
        emap = {'KeystonePublic': {'uri': 'http://foo:8000/'}}
        stack.to_dict.return_value = {
            'outputs': [{'output_key': 'EndpointMap',
                         'output_value': emap}]
        }

        endpoint_map = utils.get_stack_output_item(stack, 'EndpointMap')
        self.assertEqual(endpoint_map,
                         {'KeystonePublic': {'uri': 'http://foo:8000/'}})

    def test_get_stack_output_item_not_found(self):
        stack = mock.MagicMock()
        stack.to_dict.return_value = {
            'outputs': [{'output_key': 'foo',
                         'output_value': 'bar'}]
        }

        val = utils.get_stack_output_item(stack, 'baz')
        self.assertEqual(val, None)

    def test_get_stack_output_item_no_stack(self):
        stack = None
        val = utils.get_stack_output_item(stack, 'baz')
        self.assertEqual(val, None)


class TestGetEndpointMap(TestCase):

    def test_get_endpoint_map(self):
        stack = mock.MagicMock()
        emap = {'KeystonePublic': {'uri': 'http://foo:8000/'}}
        stack.to_dict.return_value = {
            'outputs': [{'output_key': 'EndpointMap',
                         'output_value': emap}]
        }

        endpoint_map = utils.get_endpoint_map(stack)
        self.assertEqual(endpoint_map,
                         {'KeystonePublic': {'uri': 'http://foo:8000/'}})


class TestNodeGetCapabilities(TestCase):
    def test_with_capabilities(self):
        node = mock.Mock(properties={'capabilities': 'x:y,foo:bar'})
        self.assertEqual({'x': 'y', 'foo': 'bar'},
                         utils.node_get_capabilities(node))

    def test_no_capabilities(self):
        node = mock.Mock(properties={})
        self.assertEqual({}, utils.node_get_capabilities(node))


class TestNodeAddCapabilities(TestCase):
    def test_add(self):
        bm_client = mock.Mock()
        node = mock.Mock(uuid='uuid1', properties={})
        new_caps = utils.node_add_capabilities(bm_client, node, x='y')
        bm_client.node.update.assert_called_once_with(
            'uuid1', [{'op': 'add', 'path': '/properties/capabilities',
                       'value': 'x:y'}])
        self.assertEqual('x:y', node.properties['capabilities'])
        self.assertEqual({'x': 'y'}, new_caps)


class FakeFlavor(object):
    def __init__(self, name, profile=''):
        self.name = name
        self.profile = name
        if profile != '':
            self.profile = profile

    def get_keys(self):
        return {
            'capabilities:boot_option': 'local',
            'capabilities:profile': self.profile
        }


class TestAssignVerifyProfiles(TestCase):
    def setUp(self):

        super(TestAssignVerifyProfiles, self).setUp()
        self.bm_client = mock.Mock(spec=['node'],
                                   node=mock.Mock(spec=['list', 'update']))
        self.nodes = []
        self.bm_client.node.list.return_value = self.nodes
        self.flavors = {name: (FakeFlavor(name), 1)
                        for name in ('compute', 'control')}

    def _get_fake_node(self, profile=None, possible_profiles=[],
                       provision_state='available'):
        caps = {'%s_profile' % p: '1'
                for p in possible_profiles}
        if profile is not None:
            caps['profile'] = profile
        caps = utils.dict_to_capabilities(caps)
        return mock.Mock(uuid=str(uuid4()),
                         properties={'capabilities': caps},
                         provision_state=provision_state,
                         spec=['uuid', 'properties', 'provision_state'])

    def _test(self, expected_errors, expected_warnings,
              assign_profiles=True, dry_run=False):
        errors, warnings = utils.assign_and_verify_profiles(self.bm_client,
                                                            self.flavors,
                                                            assign_profiles,
                                                            dry_run)
        self.assertEqual(errors, expected_errors)
        self.assertEqual(warnings, expected_warnings)

    def test_no_matching_without_scale(self):
        self.flavors = {name: (object(), 0)
                        for name in self.flavors}
        self.nodes[:] = [self._get_fake_node(profile='fake'),
                         self._get_fake_node(profile='fake')]

        self._test(0, 0)
        self.assertFalse(self.bm_client.node.update.called)

    def test_exact_match(self):
        self.nodes[:] = [self._get_fake_node(profile='compute'),
                         self._get_fake_node(profile='control')]

        self._test(0, 0)
        self.assertFalse(self.bm_client.node.update.called)

    def test_nodes_with_no_profiles_present(self):
        self.nodes[:] = [self._get_fake_node(profile='compute'),
                         self._get_fake_node(profile=None),
                         self._get_fake_node(profile='foobar'),
                         self._get_fake_node(profile='control')]

        self._test(0, 1)
        self.assertFalse(self.bm_client.node.update.called)

    def test_more_nodes_with_profiles_present(self):
        self.nodes[:] = [self._get_fake_node(profile='compute'),
                         self._get_fake_node(profile='compute'),
                         self._get_fake_node(profile='compute'),
                         self._get_fake_node(profile='control')]

        self._test(0, 1)
        self.assertFalse(self.bm_client.node.update.called)

    def test_no_nodes(self):
        # One error per each flavor
        self._test(2, 0)
        self.assertFalse(self.bm_client.node.update.called)

    def test_not_enough_nodes(self):
        self.nodes[:] = [self._get_fake_node(profile='compute')]
        self._test(1, 0)
        self.assertFalse(self.bm_client.node.update.called)

    def test_assign_profiles(self):
        self.nodes[:] = [self._get_fake_node(possible_profiles=['compute']),
                         self._get_fake_node(possible_profiles=['control']),
                         self._get_fake_node(possible_profiles=['compute'])]

        # one warning for a redundant node
        self._test(0, 1, assign_profiles=True)
        self.assertEqual(2, self.bm_client.node.update.call_count)

        actual_profiles = [utils.node_get_capabilities(node).get('profile')
                           for node in self.nodes]
        actual_profiles.sort(key=lambda x: str(x))
        self.assertEqual([None, 'compute', 'control'], actual_profiles)

    def test_assign_profiles_multiple_options(self):
        self.nodes[:] = [self._get_fake_node(possible_profiles=['compute',
                                                                'control']),
                         self._get_fake_node(possible_profiles=['compute',
                                                                'control'])]

        self._test(0, 0, assign_profiles=True)
        self.assertEqual(2, self.bm_client.node.update.call_count)

        actual_profiles = [utils.node_get_capabilities(node).get('profile')
                           for node in self.nodes]
        actual_profiles.sort(key=lambda x: str(x))
        self.assertEqual(['compute', 'control'], actual_profiles)

    def test_assign_profiles_not_enough(self):
        self.nodes[:] = [self._get_fake_node(possible_profiles=['compute']),
                         self._get_fake_node(possible_profiles=['compute']),
                         self._get_fake_node(possible_profiles=['compute'])]

        self._test(1, 1, assign_profiles=True)
        # no node update for failed flavor
        self.assertEqual(1, self.bm_client.node.update.call_count)

        actual_profiles = [utils.node_get_capabilities(node).get('profile')
                           for node in self.nodes]
        actual_profiles.sort(key=lambda x: str(x))
        self.assertEqual([None, None, 'compute'], actual_profiles)

    def test_assign_profiles_dry_run(self):
        self.nodes[:] = [self._get_fake_node(possible_profiles=['compute']),
                         self._get_fake_node(possible_profiles=['control']),
                         self._get_fake_node(possible_profiles=['compute'])]

        self._test(0, 1, dry_run=True)
        self.assertFalse(self.bm_client.node.update.called)

        actual_profiles = [utils.node_get_capabilities(node).get('profile')
                           for node in self.nodes]
        self.assertEqual([None] * 3, actual_profiles)

    def test_scale(self):
        # active nodes with assigned profiles are fine
        self.nodes[:] = [self._get_fake_node(profile='compute',
                                             provision_state='active'),
                         self._get_fake_node(profile='control')]

        self._test(0, 0, assign_profiles=True)
        self.assertFalse(self.bm_client.node.update.called)

    def test_assign_profiles_wrong_state(self):
        # active nodes are not considered for assigning profiles
        self.nodes[:] = [self._get_fake_node(possible_profiles=['compute'],
                                             provision_state='active'),
                         self._get_fake_node(possible_profiles=['control'],
                                             provision_state='cleaning'),
                         self._get_fake_node(profile='compute',
                                             provision_state='error')]

        self._test(2, 1, assign_profiles=True)
        self.assertFalse(self.bm_client.node.update.called)

    def test_no_spurious_warnings(self):
        self.nodes[:] = [self._get_fake_node(profile=None)]
        self.flavors = {'baremetal': (FakeFlavor('baremetal', None), 1)}
        self._test(0, 0)


class TestPromptUser(TestCase):
    def setUp(self):
        super(TestPromptUser, self).setUp()
        self.logger = mock.MagicMock()
        self.logger.info = mock.MagicMock()

    @mock.patch('sys.stdin')
    def test_user_accepts(self, stdin_mock):
        stdin_mock.isatty.return_value = True
        stdin_mock.readline.return_value = "yes"
        result = utils.prompt_user_for_confirmation("[y/N]?", self.logger)
        self.assertTrue(result)

    @mock.patch('sys.stdin')
    def test_user_declines(self, stdin_mock):
        stdin_mock.isatty.return_value = True
        stdin_mock.readline.return_value = "no"
        result = utils.prompt_user_for_confirmation("[y/N]?", self.logger)
        self.assertFalse(result)

    @mock.patch('sys.stdin')
    def test_user_no_tty(self, stdin_mock):
        stdin_mock.isatty.return_value = False
        stdin_mock.readline.return_value = "yes"
        result = utils.prompt_user_for_confirmation("[y/N]?", self.logger)
        self.assertFalse(result)

    @mock.patch('sys.stdin')
    def test_user_aborts_control_c(self, stdin_mock):
        stdin_mock.isatty.return_value = False
        stdin_mock.readline.side_effect = KeyboardInterrupt()
        result = utils.prompt_user_for_confirmation("[y/N]?", self.logger)
        self.assertFalse(result)

    @mock.patch('sys.stdin')
    def test_user_aborts_with_control_d(self, stdin_mock):
        stdin_mock.isatty.return_value = False
        stdin_mock.readline.side_effect = EOFError()
        result = utils.prompt_user_for_confirmation("[y/N]?", self.logger)
        self.assertFalse(result)


class TestReplaceLinks(TestCase):

    def setUp(self):
        super(TestReplaceLinks, self).setUp()
        self.link_replacement = {
            'file:///home/stack/test.sh':
                'user-files/home/stack/test.sh',
            'file:///usr/share/extra-templates/my.yml':
                'user-files/usr/share/extra-templates/my.yml',
        }

    def test_replace_links(self):
        source = (
            'description: my template\n'
            'heat_template_version: "2014-10-16"\n'
            'parameters:\n'
            '  foo:\n'
            '    default: ["bar"]\n'
            '    type: json\n'
            '  bar:\n'
            '    default: []\n'
            'resources:\n'
            '  test_config:\n'
            '    properties:\n'
            '      config: {get_file: "file:///home/stack/test.sh"}\n'
            '    type: OS::Heat::SoftwareConfig\n'
        )
        expected = (
            'description: my template\n'
            'heat_template_version: "2014-10-16"\n'
            'parameters:\n'
            '  foo:\n'
            '    default: ["bar"]\n'
            '    type: json\n'
            '  bar:\n'
            '    default: []\n'
            'resources:\n'
            '  test_config:\n'
            '    properties:\n'
            '      config: {get_file: user-files/home/stack/test.sh}\n'
            '    type: OS::Heat::SoftwareConfig\n'
        )

        # the yaml->string dumps aren't always character-precise, so
        # we need to parse them into dicts for comparison
        expected_dict = yaml.safe_load(expected)
        result_dict = yaml.safe_load(utils.replace_links_in_template_contents(
            source, self.link_replacement))
        self.assertEqual(expected_dict, result_dict)

    def test_replace_links_not_template(self):
        # valid JSON/YAML, but doesn't have heat_template_version
        source = '{"get_file": "file:///home/stack/test.sh"}'
        self.assertEqual(
            source,
            utils.replace_links_in_template_contents(
                source, self.link_replacement))

    def test_replace_links_not_yaml(self):
        # invalid JSON/YAML -- curly brace left open
        source = '{"invalid JSON"'
        self.assertEqual(
            source,
            utils.replace_links_in_template_contents(
                source, self.link_replacement))

    def test_relative_link_replacement(self):
        current_dir = 'user-files/home/stack'
        expected = {
            'file:///home/stack/test.sh':
                'test.sh',
            'file:///usr/share/extra-templates/my.yml':
                '../../usr/share/extra-templates/my.yml',
        }
        self.assertEqual(expected, utils.relative_link_replacement(
            self.link_replacement, current_dir))


class TestBracketIPV6(TestCase):
    def test_basic(self):
        result = utils.bracket_ipv6('::1')
        self.assertEqual('[::1]', result)

    def test_hostname(self):
        result = utils.bracket_ipv6('hostname')
        self.assertEqual('hostname', result)

    def test_already_bracketed(self):
        result = utils.bracket_ipv6('[::1]')
        self.assertEqual('[::1]', result)


class TestIsValidIP(TestCase):
    def test_with_valid_ipv4(self):
        result = utils.is_valid_ip('192.168.0.1')
        self.assertEqual(True, result)

    def test_with_valid_ipv6(self):
        result = utils.is_valid_ip('::1')
        self.assertEqual(True, result)

    def test_with_invalid_ip(self):
        result = utils.is_valid_ip('192.168.1%bad')
        self.assertEqual(False, result)


class TestIsLoopback(TestCase):
    def test_with_loopback(self):
        result = utils.is_loopback('127.0.0.1')
        self.assertEqual(True, result)

    def test_with_no_loopback(self):
        result = utils.is_loopback('10.0.0.1')
        self.assertEqual(False, result)


class TestGetHostIps(TestCase):
    def test_get_host_ips(self):
        with mock.patch.object(socket, 'getaddrinfo') as mock_addrinfo:
            mock_addrinfo.return_value = [('', '', 6, '', ('127.0.0.1', 0))]
            result = utils.get_host_ips('myhost.domain')
            self.assertEqual(['127.0.0.1'], result)


class TestGetSingleIp(TestCase):
    def test_with_fqdn_and_valid_ip(self):
        with mock.patch.object(utils, 'get_host_ips') as mock_gethostips:
            mock_gethostips.return_value = ['192.168.0.1']
            result = utils.get_single_ip('myhost.domain')
            self.assertEqual('192.168.0.1', result)

    def test_with_fqdn_and_loopback(self):
        with mock.patch.object(utils, 'get_host_ips') as mock_gethostips:
            mock_gethostips.return_value = ['127.0.0.1']
            self.assertRaises(exceptions.LookupError,
                              utils.get_single_ip, 'myhost.domain')

    def test_with_too_much_ips(self):
        with mock.patch.object(utils, 'get_host_ips') as mock_gethostips:
            mock_gethostips.return_value = ['192.168.0.1', '192.168.0.2']
            self.assertRaises(exceptions.LookupError,
                              utils.get_single_ip, 'myhost.domain')

    def test_without_ip(self):
        with mock.patch.object(utils, 'get_host_ips') as mock_gethostips:
            mock_gethostips.return_value = []
            self.assertRaises(exceptions.LookupError,
                              utils.get_single_ip, 'myhost.domain')

    def test_with_invalid_ip(self):
        with mock.patch.object(utils, 'get_host_ips') as mock_gethostips:
            mock_gethostips.return_value = ['192.168.23.x']
            self.assertRaises(exceptions.LookupError,
                              utils.get_single_ip, 'myhost.domain')


class TestStoreCliParam(TestCase):

    def setUp(self):
        self.args = argparse.ArgumentParser()

    @mock.patch('os.path.isdir')
    @mock.patch('os.path.exists')
    def test_exists_but_not_dir(self, mock_exists, mock_isdir):
        mock_exists.return_value = True
        mock_isdir.return_value = False
        self.assertRaises(exceptions.InvalidConfiguration,
                          utils.store_cli_param,
                          "overcloud deploy", self.args)

    @mock.patch('os.path.isdir')
    @mock.patch('os.path.exists')
    def test_write_cli_param(self, mock_exists, mock_isdir):
        history_path = os.path.join(os.path.expanduser("~"), '.tripleo')
        mock_exists.return_value = True
        mock_isdir.return_value = True
        mock_file = mock.mock_open()

        class ArgsFake(object):
            def __init__(self):
                self.a = 1

        dt = datetime.datetime(2017, 11, 22)
        with mock.patch("six.moves.builtins.open", mock_file):
            with mock.patch('tripleoclient.utils.datetime') as mock_date:
                mock_date.datetime.now.return_value = dt
                utils.store_cli_param("overcloud plan list", ArgsFake())

        expected_call = [
            mock.call("%s/history" % history_path, 'a'),
            mock.call().write('2017-11-22 00:00:00 overcloud-plan-list a=1 \n')
        ]
        mock_file.assert_has_calls(expected_call, any_order=True)

    @mock.patch('six.moves.builtins.open')
    @mock.patch('os.path.isdir')
    @mock.patch('os.path.exists')
    def test_fail_to_write_data(self, mock_exists, mock_isdir, mock_open):
        mock_exists.return_value = True
        mock_isdir.return_value = True
        mock_open.side_effect = IOError()
        self.assertRaises(IOError, utils.store_cli_param, "command", self.args)


class ProcessMultipleEnvironments(TestCase):

    def setUp(self):
        self.tht_root = '/twd/templates'
        self.user_tht_root = '/tmp/thtroot/'
        self.created_env_files = [
            './inside.yaml', '/tmp/thtroot/abs.yaml',
            '/tmp/thtroot/puppet/foo.yaml',
            '/tmp/thtroot/environments/myenv.yaml',
            '/tmp/thtroot42/notouch.yaml',
            './tmp/thtroot/notouch2.yaml',
            '../outside.yaml']

    @mock.patch('heatclient.common.template_utils.'
                'process_environment_and_files', return_value=({}, {}),
                autospec=True)
    @mock.patch('heatclient.common.template_utils.'
                'get_template_contents', return_value=({}, {}),
                autospec=True)
    @mock.patch('heatclient.common.environment_format.'
                'parse', autospec=True, return_value=dict())
    @mock.patch('heatclient.common.template_format.'
                'parse', autospec=True, return_value=dict())
    def test_redirect_templates_paths(self,
                                      mock_hc_templ_parse,
                                      mock_hc_env_parse,
                                      mock_hc_get_templ_cont,
                                      mock_hc_process):
        utils.process_multiple_environments(self.created_env_files,
                                            self.tht_root,
                                            self.user_tht_root)

        mock_hc_process.assert_has_calls([
            mock.call(env_path='./inside.yaml'),
            mock.call(env_path='/twd/templates/abs.yaml'),
            mock.call(env_path='/twd/templates/puppet/foo.yaml'),
            mock.call(env_path='/twd/templates/environments/myenv.yaml'),
            mock.call(env_path='/tmp/thtroot42/notouch.yaml'),
            mock.call(env_path='./tmp/thtroot/notouch2.yaml'),
            mock.call(env_path='../outside.yaml')])

    @mock.patch('heatclient.common.template_utils.'
                'process_environment_and_files', return_value=({}, {}),
                autospec=True)
    @mock.patch('heatclient.common.template_utils.'
                'get_template_contents', return_value=({}, {}),
                autospec=True)
    @mock.patch('heatclient.common.environment_format.'
                'parse', autospec=True, return_value=dict())
    @mock.patch('heatclient.common.template_format.'
                'parse', autospec=True, return_value=dict())
    @mock.patch('yaml.safe_dump', autospec=True)
    @mock.patch('yaml.safe_load', autospec=True)
    @mock.patch('six.moves.builtins.open')
    @mock.patch('tempfile.NamedTemporaryFile', autospec=True)
    def test_rewrite_env_files(self,
                               mock_temp, mock_open,
                               mock_yaml_load,
                               mock_yaml_dump,
                               mock_hc_templ_parse,
                               mock_hc_env_parse,
                               mock_hc_get_templ_cont,
                               mock_hc_process):

        def hc_process(*args, **kwargs):
            if 'abs.yaml' in kwargs['env_path']:
                raise hc_exc.CommandError
            else:
                return ({}, {})

        mock_hc_process.side_effect = hc_process
        rewritten_env = {'resource_registry': {
            'OS::Foo::Bar': '/twd/outside.yaml',
            'OS::Foo::Baz': '/twd/templates/inside.yaml',
            'OS::Foo::Qux': '/twd/templates/abs.yaml',
            'OS::Foo::Quux': '/tmp/thtroot42/notouch.yaml',
            'OS::Foo::Corge': '/twd/templates/puppet/foo.yaml'
            }
        }
        myenv = {'resource_registry': {
            'OS::Foo::Bar': '../outside.yaml',
            'OS::Foo::Baz': './inside.yaml',
            'OS::Foo::Qux': '/tmp/thtroot/abs.yaml',
            'OS::Foo::Quux': '/tmp/thtroot42/notouch.yaml',
            'OS::Foo::Corge': '/tmp/thtroot/puppet/foo.yaml'
            }
        }
        mock_yaml_load.return_value = myenv

        utils.process_multiple_environments(self.created_env_files,
                                            self.tht_root,
                                            self.user_tht_root, False)

        mock_yaml_dump.assert_has_calls([mock.call(rewritten_env,
                                        default_flow_style=False)])


class GetTripleoAnsibleInventory(TestCase):

    def setUp(self):
        super(GetTripleoAnsibleInventory, self).setUp()
        self.inventory_file = ''
        self.ssh_user = 'heat_admin'
        self.stack = 'foo-overcloud'

    @mock.patch('tripleoclient.utils.get_tripleo_ansible_inventory',
                autospec=True)
    def test_get_tripleo_ansible_inventory(self, mock_inventory):

        with mock.patch('os.path.exists') as mock_exists:
            mock_exists.return_value = True

            self.cmd = utils.get_tripleo_ansible_inventory(
                inventory_file=self.inventory_file,
                ssh_user=self.ssh_user,
                stack=self.stack)

            self.cmd.take_action()

            mock_inventory.assert_called_once_with(
                inventory_file='',
                ssh_user='heat_admin',
                stack='foo-overcloud'
            )


class TestNormalizeFilePath(TestCase):

    @mock.patch('os.path.isfile', return_value=True)
    def test_norm_path_abs(self, mock_exists):
        self.assertEqual(
            utils.rel_or_abs_path('/foobar.yaml', '/tmp'),
            '/foobar.yaml')

    @mock.patch('os.path.isfile', side_effect=[False, True])
    def test_norm_path_rel(self, mock_exists):
        self.assertEqual(
            utils.rel_or_abs_path('baz/foobar.yaml', '/bar'),
            '/bar/baz/foobar.yaml')


class TestFetchRolesFile(TestCase):

    @mock.patch('os.path.exists', return_value=True)
    def test_fetch_roles_file(self, mock_exists):
        with tempfile.NamedTemporaryFile(mode='w') as roles_file:
            yaml.dump([{'name': 'Foobar'}], roles_file)
            with mock.patch('tripleoclient.utils.rel_or_abs_path') as mock_rf:
                mock_rf.return_value = roles_file.name
                self.assertEqual(utils.fetch_roles_file(roles_file.name),
                                 [{'name': 'Foobar'}])


class TestOvercloudNameScenarios(TestWithScenarios):
    scenarios = [
        ('kernel_default',
         dict(func=utils.overcloud_kernel,
              basename='overcloud-full',
              expected=('overcloud-full-vmlinuz', '.vmlinuz'))),
        ('kernel_arch',
         dict(func=utils.overcloud_kernel,
              basename='overcloud-full',
              arch='x86_64',
              expected=('x86_64-overcloud-full-vmlinuz', '.vmlinuz'))),
        ('kernel_arch_platform',
         dict(func=utils.overcloud_kernel,
              basename='overcloud-full',
              arch='x86_64',
              platform='SNB',
              expected=('SNB-x86_64-overcloud-full-vmlinuz', '.vmlinuz'))),
        ('kernel_platform',
         dict(func=utils.overcloud_kernel,
              basename='overcloud-full',
              platform='SNB',
              expected=('overcloud-full-vmlinuz', '.vmlinuz'))),
        ('ramdisk_default',
         dict(func=utils.overcloud_ramdisk,
              basename='overcloud-full',
              expected=('overcloud-full-initrd', '.initrd'))),
        ('ramdisk_arch',
         dict(func=utils.overcloud_ramdisk,
              basename='overcloud-full',
              arch='x86_64',
              expected=('x86_64-overcloud-full-initrd', '.initrd'))),
        ('ramdisk_arch_platform',
         dict(func=utils.overcloud_ramdisk,
              basename='overcloud-full',
              arch='x86_64',
              platform='SNB',
              expected=('SNB-x86_64-overcloud-full-initrd', '.initrd'))),
        ('ramdisk_platform',
         dict(func=utils.overcloud_ramdisk,
              basename='overcloud-full',
              platform='SNB',
              expected=('overcloud-full-initrd', '.initrd'))),
        ('image_default',
         dict(func=utils.overcloud_image,
              basename='overcloud-full',
              expected=('overcloud-full', '.raw'))),
        ('image_arch',
         dict(func=utils.overcloud_image,
              basename='overcloud-full',
              arch='x86_64',
              expected=('x86_64-overcloud-full', '.raw'))),
        ('image_arch_platform',
         dict(func=utils.overcloud_image,
              basename='overcloud-full',
              arch='x86_64',
              platform='SNB',
              expected=('SNB-x86_64-overcloud-full', '.raw'))),
        ('image_platform',
         dict(func=utils.overcloud_image,
              basename='overcloud-full',
              platform='SNB',
              expected=('overcloud-full', '.raw'))),
    ]

    def test_overcloud_params(self):
        kwargs = dict()
        for attr in ['arch', 'platform']:
            if hasattr(self, attr):
                kwargs[attr] = getattr(self, attr)

        if kwargs:
            observed = self.func(self.basename, **kwargs)
        else:
            observed = self.func(self.basename)

        self.assertEqual(self.expected, observed)


class TestDeployNameScenarios(TestWithScenarios):
    scenarios = [
        ('kernel_default',
         dict(func=utils.deploy_kernel,
              expected='agent.kernel')),
        ('kernel_arch',
         dict(func=utils.deploy_kernel,
              arch='x86_64',
              expected='x86_64/agent.kernel')),
        ('kernel_arch_platform',
         dict(func=utils.deploy_kernel,
              arch='x86_64',
              platform='SNB',
              expected='SNB-x86_64/agent.kernel')),
        ('kernel_platform',
         dict(func=utils.deploy_kernel,
              platform='SNB',
              expected='agent.kernel')),
        ('ramdisk_default',
         dict(func=utils.deploy_ramdisk,
              expected='agent.ramdisk')),
        ('ramdisk_arch',
         dict(func=utils.deploy_ramdisk,
              arch='x86_64',
              expected='x86_64/agent.ramdisk')),
        ('ramdisk_arch_platform',
         dict(func=utils.deploy_ramdisk,
              arch='x86_64',
              platform='SNB',
              expected='SNB-x86_64/agent.ramdisk')),
        ('ramdisk_platform',
         dict(func=utils.deploy_ramdisk,
              platform='SNB',
              expected='agent.ramdisk')),
    ]

    def test_deploy_params(self):
        kwargs = {}
        for attr in ['arch', 'platform']:
            if hasattr(self, attr):
                kwargs[attr] = getattr(self, attr)

        if kwargs:
            observed = self.func(**kwargs)
        else:
            observed = self.func()

        self.assertEqual(self.expected, observed)


class TestDeploymentPythonInterpreter(TestCase):
    def test_system_default(self):
        args = mock.MagicMock()
        args.deployment_python_interpreter = None
        py = utils.get_deployment_python_interpreter(args)
        self.assertEqual(py, sys.executable)

    def test_provided_interpreter(self):
        args = mock.MagicMock()
        args.deployment_python_interpreter = 'foo'
        py = utils.get_deployment_python_interpreter(args)
        self.assertEqual(py, 'foo')


class TestWaitApiPortReady(TestCase):
    @mock.patch('six.moves.urllib.request.urlopen')
    def test_success(self, urlopen_mock):
        has_errors = utils.wait_api_port_ready(8080)
        self.assertFalse(has_errors)

    @mock.patch(
        'six.moves.urllib.request.urlopen',
        side_effect=[
            url_error.HTTPError("", 201, None, None, None), socket.timeout,
            url_error.URLError("")
        ] * 10)
    @mock.patch('time.sleep')
    def test_throw_exception_at_max_retries(self, urlopen_mock, sleep_mock):
        with self.assertRaises(RuntimeError):
            utils.wait_api_port_ready(8080)
        self.assertEqual(urlopen_mock.call_count, 30)
        self.assertEqual(sleep_mock.call_count, 30)

    @mock.patch(
        'six.moves.urllib.request.urlopen',
        side_effect=[
            socket.timeout,
            url_error.URLError(""),
            url_error.HTTPError("", 201, None, None, None), None
        ])
    @mock.patch('time.sleep')
    def test_recovers_from_exception(self, urlopen_mock, sleep_mock):
        self.assertFalse(utils.wait_api_port_ready(8080))
        self.assertEqual(urlopen_mock.call_count, 4)
        self.assertEqual(sleep_mock.call_count, 4)

    @mock.patch(
        'six.moves.urllib.request.urlopen',
        side_effect=[
            socket.timeout,
            url_error.URLError(""),
            url_error.HTTPError("", 300, None, None, None)
        ] * 10)
    @mock.patch('time.sleep')
    def test_recovers_from_multiple_choices_error_code(self, urlopen_mock,
                                                       sleep_mock):
        self.assertTrue(utils.wait_api_port_ready(8080))
        self.assertEqual(urlopen_mock.call_count, 3)
        self.assertEqual(sleep_mock.call_count, 3)

    @mock.patch('six.moves.urllib.request.urlopen', side_effect=NameError)
    @mock.patch('time.sleep')
    def test_dont_retry_at_unknown_exception(self, urlopen_mock, sleep_mock):
        with self.assertRaises(NameError):
            utils.wait_api_port_ready(8080)
        self.assertEqual(urlopen_mock.call_count, 1)
        self.assertEqual(sleep_mock.call_count, 1)


class TestCheckHostname(TestCase):
    @mock.patch('tripleoclient.utils.run_command')
    def test_hostname_ok(self, mock_run):
        mock_run.side_effect = ['host.domain', 'host.domain']
        mock_open_ctx = mock.mock_open(read_data='127.0.0.1 host.domain')
        with mock.patch('tripleoclient.utils.open', mock_open_ctx):
            utils.check_hostname(False)
        run_calls = [
            mock.call(['hostnamectl', '--static'], name='hostnamectl'),
            mock.call(['hostnamectl', '--transient'], name='hostnamectl')]
        self.assertEqual(mock_run.mock_calls, run_calls)

    @mock.patch('tripleoclient.utils.run_command')
    def test_hostname_fix_hosts_ok(self, mock_run):
        mock_run.side_effect = ['host.domain', 'host.domain', '']
        mock_open_ctx = mock.mock_open(read_data='')
        with mock.patch('tripleoclient.utils.open', mock_open_ctx):
            utils.check_hostname(True)
        sed_cmd = 'sed -i "s/127.0.0.1\\(\\s*\\)/127.0.0.1\\\\1host.domain ' \
                  'host /" /etc/hosts'
        run_calls = [
            mock.call(['hostnamectl', '--static'], name='hostnamectl'),
            mock.call(['hostnamectl', '--transient'], name='hostnamectl'),
            mock.call(['sudo', '/bin/bash', '-c', sed_cmd],
                      name='hostname-to-etc-hosts')]
        import pprint
        pprint.pprint(mock_run.mock_calls)
        self.assertEqual(mock_run.mock_calls, run_calls)

    @mock.patch('tripleoclient.utils.run_command')
    def test_hostname_mismatch_fail(self, mock_run):
        mock_run.side_effect = ['host.domain', '']
        self.assertRaises(RuntimeError, utils.check_hostname)

    @mock.patch('tripleoclient.utils.run_command')
    def test_hostname_short_fail(self, mock_run):
        mock_run.side_effect = ['host', 'host']
        self.assertRaises(RuntimeError, utils.check_hostname)


class TestCheckEnvForProxy(TestCase):
    def test_no_proxy(self):
        utils.check_env_for_proxy()

    @mock.patch.dict(os.environ,
                     {'http_proxy': 'foo:1111',
                      'no_proxy': 'foo'})
    def test_http_proxy_ok(self):
        utils.check_env_for_proxy(['foo'])

    @mock.patch.dict(os.environ,
                     {'https_proxy': 'bar:1111',
                      'no_proxy': 'foo,bar'})
    def test_https_proxy_ok(self):
        utils.check_env_for_proxy(['foo', 'bar'])

    @mock.patch.dict(os.environ,
                     {'http_proxy': 'foo:1111',
                      'https_proxy': 'bar:1111',
                      'no_proxy': 'foobar'})
    def test_proxy_fail(self):
        self.assertRaises(RuntimeError,
                          utils.check_env_for_proxy,
                          ['foo', 'bar'])

    @mock.patch.dict(os.environ,
                     {'http_proxy': 'foo:1111',
                      'https_proxy': 'bar:1111',
                      'no_proxy': 'foobar'})
    def test_proxy_fail_partial_match(self):
        self.assertRaises(RuntimeError,
                          utils.check_env_for_proxy,
                          ['foo', 'bar'])

    @mock.patch.dict(os.environ,
                     {'http_proxy': 'foo:1111',
                      'https_proxy': 'bar:1111'})
    def test_proxy_fail_no_proxy_unset(self):
        self.assertRaises(RuntimeError,
                          utils.check_env_for_proxy,
                          ['foo', 'bar'])


class TestConfigParser(TestCase):

    def setUp(self):
        self.tmp_dir = tempfile.mkdtemp()

    def tearDown(self):
        if self.tmp_dir:
            shutil.rmtree(self.tmp_dir)
            self.tmp_dir = None

    def test_get_config_value(self):
        cfg = ConfigParser()
        cfg.add_section('foo')
        cfg.set('foo', 'bar', 'baz')
        config = utils.get_from_cfg(cfg, 'bar', 'foo')
        self.assertEqual(config, 'baz')

    def test_getboolean_config_value(self):
        cfg = ConfigParser()
        cfg.add_section('foo')
        test_data_set = [
            (True, 'True'),
            (True, 'true'),
            (False, 'False'),
            (False, 'false')
        ]
        for test_data in test_data_set:
            expected_value, config_value = test_data
            cfg.set('foo', 'bar', config_value)
            obtained_value = utils.getboolean_from_cfg(cfg, 'bar', 'foo')
            self.assertEqual(obtained_value, expected_value)

    def test_getboolean_bad_config_value(self):
        cfg = ConfigParser()
        cfg.add_section('foo')
        cfg.set('foo', 'bar', 'I am not a boolean')
        self.assertRaises(exceptions.NotFound,
                          utils.getboolean_from_cfg,
                          cfg, 'bar', 'foo')

    def test_get_config_value_multiple_files(self):
        _, cfile1_name = tempfile.mkstemp(dir=self.tmp_dir, text=True)
        _, cfile2_name = tempfile.mkstemp(dir=self.tmp_dir, text=True)
        cfiles = [cfile1_name, cfile2_name]
        cfg = ConfigParser()
        cfg.add_section('foo')
        cfg.set('foo', 'bar', 'baz')
        with open(cfile1_name, 'w') as fp:
            cfg.write(fp)
        cfg.set('foo', 'bar', 'boop')
        with open(cfile2_name, 'w') as fp:
            cfg.write(fp)
        cfgs = utils.get_read_config(cfiles)
        config = utils.get_from_cfg(cfgs, 'bar', 'foo')
        self.assertEqual(config, 'boop')

    def test_get_config_value_bad_file(self):
        self.assertRaises(AttributeError,
                          utils.get_from_cfg,
                          'does-not-exist', 'bar', 'foo')


class TestGetLocalTimezone(TestCase):
    @mock.patch('tripleoclient.utils.run_command')
    def test_get_local_timezone(self, run_mock):
        run_mock.return_value = "" \
            "               Local time: Thu 2019-03-14 12:05:49 EDT\n" \
            "           Universal time: Thu 2019-03-14 16:05:49 UTC\n" \
            "                 RTC time: Thu 2019-03-14 16:15:50\n" \
            "                Time zone: America/New_York (EDT, -0400)\n" \
            "System clock synchronized: yes\n" \
            "              NTP service: active\n"\
            "          RTC in local TZ: no\n"
        self.assertEqual('America/New_York', utils.get_local_timezone())

    @mock.patch('tripleoclient.utils.run_command')
    def test_get_local_timezone_bad_timedatectl(self, run_mock):
        run_mock.return_value = "meh"
        self.assertEqual('UTC', utils.get_local_timezone())

    @mock.patch('tripleoclient.utils.run_command')
    def test_get_local_timezone_bad_timezone_line(self, run_mock):
        run_mock.return_value = "" \
            "                Time zone: "
        self.assertEqual('UTC', utils.get_local_timezone())


class TestParseExtraVars(TestCase):
    def test_simple_case_text_format(self):
        input_parameter = ['key1=val1', 'key2=val2 key3=val3']
        expected = {
            'key1': 'val1',
            'key2': 'val2',
            'key3': 'val3'
        }
        result = utils.parse_extra_vars(input_parameter)
        self.assertEqual(result, expected)

    def test_simple_case_json_format(self):
        input_parameter = ['{"key1": "val1", "key2": "val2"}']
        expected = {
            'key1': 'val1',
            'key2': 'val2'
        }
        result = utils.parse_extra_vars(input_parameter)
        self.assertEqual(result, expected)

    def test_multiple_format(self):
        input_parameter = [
            'key1=val1', 'key2=val2 key3=val3',
            '{"key4": "val4", "key5": "val5"}']
        expected = {
            'key1': 'val1',
            'key2': 'val2',
            'key3': 'val3',
            'key4': 'val4',
            'key5': 'val5'
        }
        result = utils.parse_extra_vars(input_parameter)
        self.assertEqual(result, expected)

    def test_same_key(self):
        input_parameter = [
            'key1=val1', 'key2=val2 key3=val3',
            '{"key1": "other_value", "key5": "val5"}']
        expected = {
            'key1': 'other_value',
            'key2': 'val2',
            'key3': 'val3',
            'key5': 'val5'
        }
        result = utils.parse_extra_vars(input_parameter)
        self.assertEqual(result, expected)

    def test_with_multiple_space(self):
        input_parameter = ['key1=val1', ' key2=val2   key3=val3 ']
        expected = {
            'key1': 'val1',
            'key2': 'val2',
            'key3': 'val3'
        }
        result = utils.parse_extra_vars(input_parameter)
        self.assertEqual(result, expected)

    def test_invalid_string(self):
        input_parameter = [
            'key1=val1', 'key2=val2 key3=val3',
            '{"key1": "other_value", "key5": "val5']
        self.assertRaises(
            ValueError, utils.parse_extra_vars, input_parameter)

    def test_invalid_format(self):
        input_parameter = ['key1 val1']
        self.assertRaises(
            ValueError, utils.parse_extra_vars, input_parameter)


class TestGeneralUtils(base.TestCommand):

    def setUp(self):
        super(TestGeneralUtils, self).setUp()
        self.tc = self.app.client_manager.tripleoclient = mock.Mock()
        obj = self.tc.object_store = mock.Mock()
        obj.put_object = mock.Mock()
        obj.put_container = mock.Mock()

    def test_update_deployment_status(self):
        mock_status = {
            'status_update': 'TESTING',
            'deployment_status': 'TESTING'
        }
        utils.update_deployment_status(
            self.app.client_manager,
            'overcloud',
            mock_status
        )
        self.tc.object_store.put_object.assert_called()
        self.tc.object_store.put_container.assert_called()

    def test_playbook_limit_parse(self):
        limit_nodes = 'controller0, compute0:compute1,!compute2'
        limit_hosts_expected = 'controller0:compute0:compute1:!compute2'
        limit_hosts_actual = utils.playbook_limit_parse(limit_nodes)
        self.assertEqual(limit_hosts_actual, limit_hosts_expected)



#   Copyright 2019 Red Hat, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#
import os

from io import StringIO
import mock
import six
from unittest import TestCase
import yaml

from tripleoclient import export


class TestExport(TestCase):
    def setUp(self):
        self.unlink_patch = mock.patch('os.unlink')
        self.addCleanup(self.unlink_patch.stop)
        self.unlink_patch.start()
        self.mock_log = mock.Mock('logging.getLogger')

        outputs = [
            {'output_key': 'EndpointMap',
             'output_value': dict(em_key='em_value')},
            {'output_key': 'HostsEntry',
             'output_value': 'hosts entry'},
            {'output_key': 'GlobalConfig',
             'output_value': dict(gc_key='gc_value')},
        ]
        self.mock_stack = mock.Mock()
        self.mock_stack.to_dict.return_value = dict(outputs=outputs)
        self.mock_open = mock.mock_open(read_data='{"an_key":"an_value"}')

        ceph_inv = {
            'DistributedComputeHCI': {
                'hosts': {
                    'dcn0-distributedcomputehci-0': {
                        'storage_ip': '192.168.24.42'
                    }
                }
            },
            'mons': {
                'children': {
                    'DistributedComputeHCI': {}
                }
            }
        }
        self.mock_open_ceph_inv = mock.mock_open(read_data=str(ceph_inv))

        ceph_all = {
            'cluster': 'dcn0',
            'fsid': 'a5a22d37-e01f-4fa0-a440-c72585c7487f',
            'keys': [
                {'name': 'client.openstack'}
            ]
        }
        self.mock_open_ceph_all = mock.mock_open(read_data=str(ceph_all))

    @mock.patch('tripleoclient.utils.get_stack')
    def test_export_stack(self, mock_get_stack):
        heat = mock.Mock()
        mock_get_stack.return_value = self.mock_stack
        with mock.patch('six.moves.builtins.open', self.mock_open):
            data = export.export_stack(heat, "overcloud")

        expected = \
            {'AllNodesExtraMapData': {u'an_key': u'an_value'},
             'EndpointMapOverride': {'em_key': 'em_value'},
             'ExtraHostFileEntries': 'hosts entry',
             'GlobalConfigExtraMapData': {'gc_key': 'gc_value'}}

        self.assertEqual(expected, data)
        self.mock_open.assert_called_once_with(
            os.path.join(
                os.environ.get('HOME'),
                'config-download/overcloud/group_vars/overcloud.json'),
            'r')

    @mock.patch('tripleoclient.utils.get_stack')
    def test_export_stack_should_filter(self, mock_get_stack):
        heat = mock.Mock()
        mock_get_stack.return_value = self.mock_stack
        self.mock_open = mock.mock_open(
            read_data='{"an_key":"an_value","ovn_dbs_vip":"vip"}')
        with mock.patch('six.moves.builtins.open', self.mock_open):
            data = export.export_stack(heat, "overcloud", should_filter=True)

        expected = \
            {'AllNodesExtraMapData': {u'ovn_dbs_vip': u'vip'},
             'EndpointMapOverride': {'em_key': 'em_value'},
             'ExtraHostFileEntries': 'hosts entry',
             'GlobalConfigExtraMapData': {'gc_key': 'gc_value'}}

        self.assertEqual(expected, data)
        self.mock_open.assert_called_once_with(
            os.path.join(
                os.environ.get('HOME'),
                'config-download/overcloud/group_vars/overcloud.json'),
            'r')

    @mock.patch('tripleoclient.utils.get_stack')
    def test_export_stack_cd_dir(self, mock_get_stack):
        heat = mock.Mock()
        mock_get_stack.return_value = self.mock_stack
        with mock.patch('six.moves.builtins.open', self.mock_open):
            export.export_stack(heat, "overcloud",
                                config_download_dir='/foo')
        self.mock_open.assert_called_once_with(
            '/foo/overcloud/group_vars/overcloud.json', 'r')

    @mock.patch('tripleoclient.utils.get_stack')
    def test_export_stack_stack_name(self, mock_get_stack):
        heat = mock.Mock()
        mock_get_stack.return_value = self.mock_stack
        with mock.patch('six.moves.builtins.open', self.mock_open):
            export.export_stack(heat, "control")
        mock_get_stack.assert_called_once_with(heat, 'control')

    def test_export_passwords(self):
        swift = mock.Mock()
        mock_passwords = {
            'parameter_defaults': {
                'AdminPassword': 'a_user'
            },
            'passwords': {
                'AdminPassword': 'A',
                'RpcPassword': 'B'
            }
        }
        sio = StringIO()
        sio.write(six.text_type(yaml.dump(mock_passwords)))
        sio.seek(0)
        swift.get_object.return_value = ("", sio)
        data = export.export_passwords(swift, 'overcloud')

        swift.get_object.assert_called_once_with(
            'overcloud', 'plan-environment.yaml')

        self.assertEqual(dict(AdminPassword='a_user',
                              RpcPassword='B'),
                         data)

    def test_export_passwords_excludes(self):
        swift = mock.Mock()
        mock_passwords = {
            'parameter_defaults': {
                'CephClientKey': 'cephkey'
            },
            'passwords': {
                'AdminPassword': 'A',
                'RpcPassword': 'B',
                'CephClientKey': 'cephkey',
                'CephClusterFSID': 'cephkey',
                'CephRgwKey': 'cephkey'
            }
        }
        sio = StringIO()
        sio.write(six.text_type(yaml.dump(mock_passwords)))
        sio.seek(0)
        swift.get_object.return_value = ("", sio)
        data = export.export_passwords(swift, 'overcloud')

        mock_passwords['passwords'].pop('CephClientKey')
        mock_passwords['passwords'].pop('CephClusterFSID')
        mock_passwords['passwords'].pop('CephRgwKey')

        self.assertEqual(mock_passwords['passwords'], data)

    def test_export_storage_ips(self):
        with mock.patch('six.moves.builtins.open', self.mock_open_ceph_inv):
            storage_ips = export.export_storage_ips('dcn0',
                                                    config_download_dir='/foo')
        self.assertEqual(storage_ips, ['192.168.24.42'])
        self.mock_open_ceph_inv.assert_called_once_with(
            '/foo/dcn0/ceph-ansible/inventory.yml', 'r')

    def test_export_ceph(self):
        expected = {
            'external_cluster_mon_ips': '192.168.24.42',
            'keys': [
                {'name': 'client.openstack'}
            ],
            'ceph_conf_overrides': {
                'client': {
                    'keyring': '/etc/ceph/dcn0.client.openstack.keyring'
                }
            },
            'cluster': 'dcn0',
            'fsid': 'a5a22d37-e01f-4fa0-a440-c72585c7487f',
            'dashboard_enabled': False
        }
        with mock.patch('six.moves.builtins.open', self.mock_open_ceph_all):
            data = export.export_ceph('dcn0', 'openstack',
                                      config_download_dir='/foo',
                                      mon_ips=['192.168.24.42'])
        self.assertEqual(data, expected)
        self.mock_open_ceph_all.assert_called_once_with(
            '/foo/dcn0/ceph-ansible/group_vars/all.yml', 'r')



#   Copyright 2020 Red Hat, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import argparse
import logging

from osc_lib.command import command
from osc_lib.i18n import _

from tripleoclient import constants
from tripleoclient import utils

LOG = logging.getLogger(__name__ + ".BackupUndercloud")


class BackupUndercloud(command.Command):
    """Backup the undercloud"""

    def get_parser(self, prog_name):
        parser = argparse.ArgumentParser(
            description=self.get_description(),
            prog=prog_name,
            add_help=False
        )

        # New flags for tripleo-ansible backup and restore role.
        parser.add_argument(
            '--init',
            default=False,
            action='store_true',
            help=_("Initialize enviornment for backup,"
                   "which will check for package install"
                   "status and configured ReaR.")
        )

        parser.add_argument(
            '--inventory',
            action='store',
            default='/home/stack/tripleo-inventory.yaml',
            help=_("Tripleo inventory file generated with"
                   "tripleo-ansible-inventory command.")
        )

        # Parameter to choose the files to backup
        parser.add_argument(
            '--add-path',
            action='append',
            default=['/home/stack/'],
            help=_("Add additional files to backup. "
                   "Defaults to: /home/stack/ "
                   "i.e. --add-path /this/is/a/folder/ "
                   " --add-path /this/is/a/texfile.txt")
        )
        parser.add_argument(
            "--exclude-path",
            default=[],
            action="append",
            help=_("Exclude path when performing the Undercloud Backup, "
                   "this option can be specified multiple times. "
                   "Defaults to: none "
                   "i.e. --exclude-path /this/is/a/folder/ "
                   " --exclude-path /this/is/a/texfile.txt")
        )
        parser.add_argument(
            '--save-swift',
            default=False,
            action='store_true',
            help=_("Save backup to swift. "
                   "Defaults to: False "
                   "Special attention should be taken that "
                   "Swift itself is backed up if you call this multiple times "
                   "the backup size will grow exponentially")
        )

        return parser

    def _run_backup_undercloud(self, parsed_args):

        if parsed_args.init is False:
            playbook = 'cli-undercloud-backup.yaml'
            skip_tags = None
        elif parsed_args.init is True:
            playbook = 'prepare-undercloud-backup.yaml'
            skip_tags = 'bar_create_recover_image'

        self._run_ansible_playbook(
                            playbook=playbook,
                            inventory=parsed_args.inventory,
                            skip_tags=skip_tags,
                            extra_vars=None
                            )

    def _legacy_backup_undercloud(self, parsed_args):
        """Legacy backup undercloud.

        This will allow for easier removal once the functionality
        is no longer needed.
        """

        merge_paths = sorted(list(set(parsed_args.add_path)))
        for exc in parsed_args.exclude_path:
            if exc in merge_paths:
                merge_paths.remove(exc)

        files_to_backup = ','.join(merge_paths)

        # Define the backup sources_path (files to backup).
        # This is a comma separated string.
        # I.e. "/this/is/a/folder/,/this/is/a/texfile.txt"
        extra_vars = {"sources_path": files_to_backup}
        if parsed_args.save_swift:
            extra_vars.update({"save_swift": True})

        LOG.debug(_('Launch the Undercloud Backup'))
        self._run_ansible_playbook(
            playbook='cli-undercloud-backup-legacy.yaml',
            inventory='localhost, ',
            skip_tags=None,
            extra_vars=extra_vars
            )

    def _run_ansible_playbook(self,
                              playbook,
                              inventory,
                              skip_tags,
                              extra_vars):
        """Run ansible playbook"""

        with utils.TempDirs() as tmp:
            utils.run_ansible_playbook(
                playbook=playbook,
                inventory=inventory,
                workdir=tmp,
                playbook_dir=constants.ANSIBLE_TRIPLEO_PLAYBOOKS,
                skip_tags=skip_tags,
                verbosity=utils.playbook_verbosity(self=self),
                extra_vars=extra_vars
            )

    def take_action(self, parsed_args):

        if len(parsed_args.add_path) > 1 or parsed_args.save_swift:

            LOG.warning("The following flags will be deprecated:"
                        "[--add-path, --exclude-path, --save-swift]")

            self._legacy_backup_undercloud(parsed_args)

        else:
            self._run_backup_undercloud(parsed_args)

        print(
            '\n'
            ' #############################################################\n'
            ' #                  Disclaimer                               #\n'
            ' # Backup verification is the End Users responsibility       #\n'
            ' # Please verify backup integrity before any possible        #\n'
            ' # disruptive actions against the Undercloud. The resulting  #\n'
            ' # backup file path will be shown on a successful execution. #\n'
            ' #                                                           #\n'
            ' # .-Stay safe and avoid future issues-.                     #\n'
            ' #############################################################\n'
        )



#   Copyright 2019 Red Hat, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import argparse
import getpass
import logging
import os

from cliff import command
from osc_lib.i18n import _

from tripleoclient.constants import DEFAULT_HEAT_CONTAINER
from tripleoclient import exceptions
from tripleoclient import heat_launcher


class LaunchHeat(command.Command):
    """Launch all-in-one Heat process and run in the foreground."""

    log = logging.getLogger(__name__ + ".Deploy")
    auth_required = False
    heat_pid = None

    def _kill_heat(self, parsed_args):
        """Tear down heat installer and temp files

        Kill the heat launcher/installer process.
        Teardown temp files created in the deployment process,
        when cleanup is requested.

        """
        if self.heat_pid:
            self.heat_launch.kill_heat(self.heat_pid)
            pid, ret = os.waitpid(self.heat_pid, 0)
            self.heat_pid = None

        return 0

    def _launch_heat(self, parsed_args):
        # we do this as root to chown config files properly for docker, etc.
        if parsed_args.heat_native is not None and \
                parsed_args.heat_native.lower() == "false":
            self.heat_launch = heat_launcher.HeatContainerLauncher(
                parsed_args.heat_api_port,
                parsed_args.heat_container_image,
                parsed_args.heat_user,
                parsed_args.heat_dir)
        else:
            self.heat_launch = heat_launcher.HeatNativeLauncher(
                parsed_args.heat_api_port,
                parsed_args.heat_container_image,
                parsed_args.heat_user,
                parsed_args.heat_dir)

        self.heat_launch.heat_db_sync()
        self.heat_launch.launch_heat()

        return 0

    def get_parser(self, prog_name):
        parser = argparse.ArgumentParser(
            description=self.get_description(),
            prog=prog_name,
            add_help=False
        )
        parser.add_argument(
            '--heat-api-port', metavar='<HEAT_API_PORT>',
            dest='heat_api_port',
            default='8006',
            help=_('Heat API port to use for the installers private'
                   ' Heat API instance. Optional. Default: 8006.)')
        )
        parser.add_argument(
            '--heat-user', metavar='<HEAT_USER>',
            dest='heat_user',
            default=getpass.getuser(),
            help=_('User to execute the non-privileged heat-all process. '
                   'Defaults to current user. '
                   'If the configuration files /etc/heat/heat.conf or '
                   '/usr/share/heat/heat-dist.conf exist, the user '
                   'must have read access to those files.')
        )
        parser.add_argument(
            '--heat-container-image', metavar='<HEAT_CONTAINER_IMAGE>',
            dest='heat_container_image',
            default=DEFAULT_HEAT_CONTAINER,
            help=_('The container image to use when launching the heat-all '
                   'process. Defaults to: {}'.format(DEFAULT_HEAT_CONTAINER))
        )
        parser.add_argument(
            '--heat-native',
            dest='heat_native',
            nargs='?',
            default=None,
            const="true",
            help=_('Execute the heat-all process natively on this host. '
                   'This option requires that the heat-all binaries '
                   'be installed locally on this machine. '
                   'This option is enabled by default which means heat-all is '
                   'executed on the host OS directly.')
        )
        parser.add_argument(
            '--kill', '-k',
            dest='kill',
            action='store_true',
            default=False,
            help=_('Kill the running heat process (if found).')
        )
        parser.add_argument(
            '--heat-dir',
            dest='heat_dir',
            action='store',
            default=os.path.join(os.getcwd(), 'heat-launcher'),
            help=_("Directory to use for file storage and logs of the "
                   "running heat process. Defaults to 'heat-launcher' "
                   "in the current directory.")
        )
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)" % parsed_args)
        if parsed_args.kill:
            if self._kill_heat(parsed_args) != 0:
                msg = _('Heat kill failed.')
                self.log.error(msg)
                raise exceptions.DeploymentError(msg)
        else:
            if self._launch_heat(parsed_args) != 0:
                msg = _('Heat launch failed.')
                self.log.error(msg)
                raise exceptions.DeploymentError(msg)



#   Copyright 2015 Red Hat, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import mock
import os
import tempfile

import yaml

from tripleoclient.tests.v1.overcloud_netenv_validate import fakes
from tripleoclient.v1 import overcloud_netenv_validate


EMPTY_NETENV = """resource_registry:
  OS::TripleO::BlockStorage::Net::SoftwareConfig: /tmp/foo

parameter_defaults:
  NeutronExternalNetworkBridge: "''"
"""


class TestValidateOvercloudNetenv(fakes.TestValidateOvercloudNetenv):

    def setUp(self):
        super(TestValidateOvercloudNetenv, self).setUp()

        # Get the command object to test
        self.cmd = overcloud_netenv_validate.ValidateOvercloudNetenv(
            self.app, None)

    def temporary_nic_config_file(self, bridges):
        nic_config = {
            'resources': {
                'OsNetConfigImpl': {
                    'properties': {
                        'config': {
                            'os_net_config': {
                                'network_config': bridges,
                            }
                        }
                    }
                }
            }
        }
        tmp = tempfile.NamedTemporaryFile(mode='w', delete=False)
        yaml.dump(nic_config, tmp)
        tmp.close()
        return tmp.name

    def test_cidr_no_overlapping_networks(self):
        networks = [
            '172.17.0.0/24',
            '172.16.0.0/24',
            '172.17.1.0/24',
            '172.17.2.0/24',
            '10.1.2.0/24',
        ]
        self.cmd.error_count = 0
        self.cmd.check_cidr_overlap(networks)
        self.assertEqual(0, self.cmd.error_count)

    def test_cidr_overlapping_networks(self):
        networks = [
            '172.17.1.0/24',
            '172.17.1.0/24',
            '10.1.2.0/24',
        ]
        self.cmd.error_count = 0
        self.cmd.check_cidr_overlap(networks)
        self.assertEqual(1, self.cmd.error_count)

    def test_cidr_nonnumerical_address(self):
        networks = [
            'nonsense',
        ]
        self.cmd.error_count = 0
        self.cmd.check_cidr_overlap(networks)
        self.assertEqual(1, self.cmd.error_count)

    def test_cidr_address_outside_of_range(self):
        networks = [
            '172.17.0.278/24',
        ]
        self.cmd.error_count = 0
        self.cmd.check_cidr_overlap(networks)
        self.assertEqual(1, self.cmd.error_count)

    def test_vlan_ids_unique(self):
        vlans = {
            'InternalApiNetworkVlanID': 201,
            'StorageNetworkVlanID': 202,
            'StorageMgmtNetworkVlanID': 203,
            'TenantNetworkVlanID': 204,
            'ExternalNetworkVlanID': 100,
        }
        self.cmd.error_count = 0
        self.cmd.check_vlan_ids(vlans)
        self.assertEqual(0, self.cmd.error_count)

    def test_vlan_ids_duplicate(self):
        vlans = {
            'InternalApiNetworkVlanID': 201,
            'StorageNetworkVlanID': 202,
            'StorageMgmtNetworkVlanID': 203,
            'TenantNetworkVlanID': 202,  # conflicts with StorageNetworkVlanID
            'ExternalNetworkVlanID': 100,
        }
        self.cmd.error_count = 0
        self.cmd.check_vlan_ids(vlans)
        self.assertEqual(1, self.cmd.error_count)

    def test_allocation_pools_pairing_no_overlap(self):
        filedata = {
            'InternalApiNetCidr': '172.17.0.0/24',
            'StorageNetCidr': '172.18.0.0/24',
            'InternalApiAllocationPools': [
                {'start': '172.17.0.10', 'end': '172.17.0.200'}],
            'StorageAllocationPools': [
                {'start': '172.18.0.10', 'end': '172.18.0.200'}],
        }
        pools = {
            'InternalApiAllocationPools': [
                {'start': '172.17.0.10', 'end': '172.17.0.200'}],
            'StorageAllocationPools': [
                {'start': '172.18.0.10', 'end': '172.18.0.200'}],
        }
        self.cmd.error_count = 0
        self.cmd.check_allocation_pools_pairing(filedata, pools)
        self.assertEqual(0, self.cmd.error_count)

    def test_allocation_pools_pairing_inverse_range(self):
        filedata = {
            'InternalApiNetCidr': '172.17.0.0/24',
            'StorageNetCidr': '172.18.0.0/24',
            'InternalApiAllocationPools': [
                {'start': '172.17.0.200', 'end': '172.17.0.10'}],
            'StorageAllocationPools': [
                {'start': '172.18.0.10', 'end': '172.18.0.200'}],
        }
        pools = {
            'InternalApiAllocationPools': [
                {'start': '172.17.0.200', 'end': '172.17.0.10'}],
            'StorageAllocationPools': [
                {'start': '172.18.0.10', 'end': '172.18.0.200'}],
        }
        self.cmd.error_count = 0
        self.cmd.check_allocation_pools_pairing(filedata, pools)
        self.assertEqual(1, self.cmd.error_count)

    def test_allocation_pools_pairing_pool_outside_subnet(self):
        filedata = {
            'InternalApiNetCidr': '172.17.0.0/24',
            'InternalApiAllocationPools': [
                {'start': '172.16.0.10', 'end': '172.16.0.200'}],
        }
        pools = {
            'InternalApiAllocationPools': [
                {'start': '172.16.0.10', 'end': '172.16.0.200'}],
        }
        self.cmd.error_count = 0
        self.cmd.check_allocation_pools_pairing(filedata, pools)
        self.assertEqual(1, self.cmd.error_count)

    def test_allocation_pools_pairing_invalid_cidr(self):
        filedata = {
            'InternalApiNetCidr': '172.17.0.298/24',
            'InternalApiAllocationPools': [
                {'start': '172.17.0.10', 'end': '172.17.0.200'}],
        }
        pools = {
            'InternalApiAllocationPools': [
                {'start': '172.17.0.10', 'end': '172.17.0.200'}],
        }
        self.cmd.error_count = 0
        self.cmd.check_allocation_pools_pairing(filedata, pools)
        self.assertEqual(1, self.cmd.error_count)

    def test_allocation_pools_pairing_invalid_range(self):
        filedata = {
            'InternalApiNetCidr': '172.17.0.0/24',
            'InternalApiAllocationPools': [
                {'start': '172.17.0.10', 'end': '172.17.0.287'}],
        }
        pools = {
            'InternalApiAllocationPools': [
                {'start': '172.17.0.10', 'end': '172.17.0.287'}],
        }
        self.cmd.error_count = 0
        self.cmd.check_allocation_pools_pairing(filedata, pools)
        self.assertEqual(1, self.cmd.error_count)

    def test_nic_nonexistent_path(self):
        self.cmd.error_count = 0
        self.cmd.NIC_validate('OS::TripleO::Controller::Net::SoftwareConfig',
                              'this file that not exist')
        self.assertEqual(1, self.cmd.error_count)

    def test_nic_valid_file(self):
        bridges = [{
            'type': 'ovs_bridge',
            'name': 'br-storage',
            'members': [
                {'type': 'interface', 'name': 'eth0'},
                {'type': 'interface', 'name': 'eth1'},
                {'type': 'ovs_bond', 'name': 'bond1'}
            ],
        }]
        tmp = self.temporary_nic_config_file(bridges)
        self.cmd.error_count = 0
        self.cmd.NIC_validate(
            'OS::TripleO::Controller::Net::SoftwareConfig', tmp)
        os.unlink(tmp)
        self.assertEqual(0, self.cmd.error_count)

    def test_nic_no_bond_too_many_interfaces(self):
        bridges = [{
            'type': 'ovs_bridge',
            'name': 'br-storage',
            'members': [
                {'type': 'interface', 'name': 'eth0'},
                {'type': 'interface', 'name': 'eth1'},
            ],
        }]
        tmp = self.temporary_nic_config_file(bridges)
        self.cmd.error_count = 0
        self.cmd.NIC_validate(
            'OS::TripleO::Controller::Net::SoftwareConfig', tmp)
        os.unlink(tmp)
        self.assertEqual(1, self.cmd.error_count)

    def test_nic_two_bonds(self):
        bridges = [{
            'type': 'ovs_bridge',
            'name': 'br-storage',
            'members': [
                {'type': 'interface', 'name': 'eth0'},
                {'type': 'interface', 'name': 'eth1'},
                {'type': 'ovs_bond', 'name': 'bond1'},
                {'type': 'ovs_bond', 'name': 'bond2'},
            ],
        }]
        tmp = self.temporary_nic_config_file(bridges)
        self.cmd.error_count = 0
        self.cmd.NIC_validate(
            'OS::TripleO::Controller::Net::SoftwareConfig', tmp)
        os.unlink(tmp)
        self.assertEqual(1, self.cmd.error_count)

    @mock.patch('tripleoclient.v1.overcloud_netenv_validate.'
                'ValidateOvercloudNetenv.NIC_validate', autospec=True)
    def test_command(self, mock_nic_validate):
        """Testing the command with a minimal file that will fail"""
        with tempfile.NamedTemporaryFile('wt') as net_file:
            net_file.write(EMPTY_NETENV)
            net_file.flush()

            arglist = ['--file', net_file.name]
            verifylist = [
                ('netenv', net_file.name),
            ]

            parsed_args = self.check_parser(self.cmd, arglist, verifylist)
            # Validating a minimal file shouldn't raise errors.
            self.cmd.take_action(parsed_args)



#   Copyright 2015 Red Hat, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import argparse
import collections
import json
import logging
import os
import sys

from cliff.formatters import table
from osc_lib import exceptions as oscexc
from osc_lib.i18n import _
from osc_lib import utils
import yaml

from tripleoclient import command
from tripleoclient import constants
from tripleoclient import utils as oooutils
from tripleoclient.workflows import baremetal

# NOTE(cloudnull): V1 imports, These classes will be removed as they're
#                  converted from mistral to ansible.
from tripleoclient.v1.overcloud_node import CleanNode  # noqa
from tripleoclient.v1.overcloud_node import ConfigureNode  # noqa
from tripleoclient.v1.overcloud_node import DeleteNode  # noqa
from tripleoclient.v1.overcloud_node import DiscoverNode  # noqa
from tripleoclient.v1.overcloud_node import ProvideNode  # noqa


class ImportNode(command.Command):
    """Import baremetal nodes from a JSON, YAML or CSV file.

    The node status will be set to 'manageable' by default.
    """

    log = logging.getLogger(__name__ + ".ImportNode")

    def get_parser(self, prog_name):
        parser = super(ImportNode, self).get_parser(prog_name)
        parser.add_argument('--introspect',
                            action='store_true',
                            help=_('Introspect the imported nodes'))
        parser.add_argument('--run-validations', action='store_true',
                            default=False,
                            help=_('Run the pre-deployment validations. These'
                                   ' external validations are from the'
                                   ' TripleO Validations project.'))
        parser.add_argument('--validate-only', action='store_true',
                            default=False,
                            help=_('Validate the env_file and then exit '
                                   'without actually importing the nodes.'))
        parser.add_argument('--provide',
                            action='store_true',
                            help=_('Provide (make available) the nodes'))
        parser.add_argument('--no-deploy-image', action='store_true',
                            help=_('Skip setting the deploy kernel and '
                                   'ramdisk.'))
        parser.add_argument('--instance-boot-option',
                            choices=['local', 'netboot'], default=None,
                            help=_('Whether to set instances for booting from'
                                   ' local hard drive (local) or network '
                                   ' (netboot).'))
        parser.add_argument("--http-boot",
                            default=os.environ.get(
                                'HTTP_BOOT',
                                constants.IRONIC_HTTP_BOOT_BIND_MOUNT),
                            help=_("Root directory for the "
                                   " ironic-python-agent image"))
        parser.add_argument('--concurrency', type=int,
                            default=20,
                            help=_('Maximum number of nodes to introspect at '
                                   'once.'))
        parser.add_argument('env_file', type=argparse.FileType('r'))
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)" % parsed_args)

        nodes_config = oooutils.parse_env_file(parsed_args.env_file)
        parsed_args.env_file.close()

        if parsed_args.validate_only:
            return baremetal.validate_nodes(self.app.client_manager,
                                            nodes_json=nodes_config)

        # Look for *specific* deploy images and update the node data if
        # one is found.
        if not parsed_args.no_deploy_image:
            oooutils.update_nodes_deploy_data(nodes_config,
                                              http_boot=parsed_args.http_boot)
        nodes = baremetal.register_or_update(
            self.app.client_manager,
            nodes_json=nodes_config,
            instance_boot_option=parsed_args.instance_boot_option
        )

        nodes_uuids = [node.uuid for node in nodes]

        if parsed_args.introspect:
            extra_vars = {
                "node_uuids": nodes_uuids,
                "run_validations": parsed_args.run_validations,
                "concurrency": parsed_args.concurrency,
            }

            with oooutils.TempDirs() as tmp:
                oooutils.run_ansible_playbook(
                    playbook='cli-baremetal-introspect.yaml',
                    inventory='localhost,',
                    workdir=tmp,
                    playbook_dir=constants.ANSIBLE_TRIPLEO_PLAYBOOKS,
                    verbosity=oooutils.playbook_verbosity(self=self),
                    extra_vars=extra_vars
                )

        if parsed_args.provide:
            baremetal.provide(
                verbosity=oooutils.playbook_verbosity(self=self),
                node_uuids=nodes_uuids
            )


class IntrospectNode(command.Command):
    """Introspect specified nodes or all nodes in 'manageable' state."""

    log = logging.getLogger(__name__ + ".IntrospectNode")

    def get_parser(self, prog_name):
        parser = super(IntrospectNode, self).get_parser(prog_name)
        group = parser.add_mutually_exclusive_group(required=True)
        group.add_argument('node_uuids',
                           nargs="*",
                           metavar="<node_uuid>",
                           default=[],
                           help=_('Baremetal Node UUIDs for the node(s) to be '
                                  'introspected'))
        group.add_argument("--all-manageable",
                           action='store_true',
                           help=_("Introspect all nodes currently in "
                                  "'manageable' state"))
        parser.add_argument('--provide',
                            action='store_true',
                            help=_('Provide (make available) the nodes once '
                                   'introspected'))
        parser.add_argument('--run-validations', action='store_true',
                            default=False,
                            help=_('Run the pre-deployment validations. These '
                                   'external validations are from the TripleO '
                                   'Validations project.'))
        parser.add_argument('--concurrency', type=int,
                            default=20,
                            help=_('Maximum number of nodes to introspect at '
                                   'once.'))
        parser.add_argument('--node-timeout', type=int,
                            default=1200,
                            help=_('Maximum timeout for node introspection.'))
        parser.add_argument('--max-retries', type=int,
                            default=1,
                            help=_('Maximum introspection retries.'))
        parser.add_argument('--retry-timeout', type=int,
                            default=120,
                            help=_('Maximum timeout between introspection'
                                   'retries'))
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)" % parsed_args)

        if parsed_args.all_manageable:
            baremetal.introspect_manageable_nodes(
                self.app.client_manager,
                run_validations=parsed_args.run_validations,
                concurrency=parsed_args.concurrency,
                node_timeout=parsed_args.node_timeout,
                max_retries=parsed_args.max_retries,
                retry_timeout=parsed_args.retry_timeout,
                verbosity=oooutils.playbook_verbosity(self=self)
            )
        else:
            baremetal.introspect(
                self.app.client_manager,
                node_uuids=parsed_args.node_uuids,
                run_validations=parsed_args.run_validations,
                concurrency=parsed_args.concurrency,
                node_timeout=parsed_args.node_timeout,
                max_retries=parsed_args.max_retries,
                retry_timeout=parsed_args.retry_timeout,
                verbosity=oooutils.playbook_verbosity(self=self)
            )

        # NOTE(cloudnull): This is using the old provide function, in a future
        #                  release this may be ported to a standalone playbook
        if parsed_args.provide:
            if parsed_args.node_uuids:
                baremetal.provide(
                    self.app.client_manager,
                    node_uuids=parsed_args.node_uuids,
                )
            else:
                baremetal.provide_manageable_nodes(
                    clients=self.app.client_manager,
                    verbosity=oooutils.playbook_verbosity(self=self)
                )


class ProvisionNode(command.Command):
    """Provision new nodes using Ironic."""

    log = logging.getLogger(__name__ + ".ProvisionNode")

    def get_parser(self, prog_name):
        parser = super(ProvisionNode, self).get_parser(prog_name)
        parser.add_argument('input',
                            metavar='<baremetal_deployment.yaml>',
                            help=_('Configuration file describing the '
                                   'baremetal deployment'))
        parser.add_argument('-o', '--output',
                            default='baremetal_environment.yaml',
                            help=_('The output environment file path'))
        parser.add_argument('--stack', dest='stack',
                            help=_('Name or ID of heat stack '
                                   '(default=Env: OVERCLOUD_STACK_NAME)'),
                            default=utils.env('OVERCLOUD_STACK_NAME',
                                              default='overcloud'))
        parser.add_argument('--overcloud-ssh-user',
                            default='heat-admin',
                            help=_('User for SSH access to newly deployed '
                                   'nodes'))
        parser.add_argument('--overcloud-ssh-key',
                            default=None,
                            help=_('Key path for ssh access to'
                                   'overcloud nodes. When undefined the key'
                                   'will be autodetected.'))
        parser.add_argument('--concurrency', type=int,
                            default=20,
                            help=_('Maximum number of nodes to provision at '
                                   'once. (default=20)'))
        parser.add_argument('--timeout', type=int,
                            default=3600,
                            help=_('Number of seconds to wait for the node '
                                   'provision to complete. (default=3600)'))
        parser.add_argument('--network-ports',
                            help=_('Enable provisioning of network ports'),
                            default=False,
                            action="store_true")
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)" % parsed_args)

        with open(parsed_args.input, 'r') as fp:
            roles = yaml.safe_load(fp)

        key = self.get_key_pair(parsed_args)
        with open('{}.pub'.format(key), 'rt') as fp:
            ssh_key = fp.read()

        output_path = os.path.abspath(parsed_args.output)

        extra_vars = {
            "stack_name": parsed_args.stack,
            "baremetal_deployment": roles,
            "baremetal_deployed_path": output_path,
            "ssh_public_keys": ssh_key,
            "ssh_user_name": parsed_args.overcloud_ssh_user,
            "node_timeout": parsed_args.timeout,
            "concurrency": parsed_args.concurrency,
            "manage_network_ports": parsed_args.network_ports
        }

        with oooutils.TempDirs() as tmp:
            oooutils.run_ansible_playbook(
                playbook='cli-overcloud-node-provision.yaml',
                inventory='localhost,',
                workdir=tmp,
                playbook_dir=constants.ANSIBLE_TRIPLEO_PLAYBOOKS,
                verbosity=oooutils.playbook_verbosity(self=self),
                extra_vars=extra_vars,
            )

        print('Nodes deployed successfully, add %s to your deployment '
              'environment' % parsed_args.output)


class UnprovisionNode(command.Command):
    """Unprovisions nodes using Ironic."""

    log = logging.getLogger(__name__ + ".UnprovisionNode")

    def get_parser(self, prog_name):
        parser = super(UnprovisionNode, self).get_parser(prog_name)
        parser.add_argument('--stack', dest='stack',
                            help=_('Name or ID of heat stack '
                                   '(default=Env: OVERCLOUD_STACK_NAME)'),
                            default=utils.env('OVERCLOUD_STACK_NAME',
                                              default='overcloud'))
        parser.add_argument('--all',
                            help=_('Unprovision every instance in the '
                                   'deployment'),
                            default=False,
                            action="store_true")
        parser.add_argument('-y', '--yes',
                            help=_('Skip yes/no prompt (assume yes)'),
                            default=False,
                            action="store_true")
        parser.add_argument('input',
                            metavar='<baremetal_deployment.yaml>',
                            help=_('Configuration file describing the '
                                   'baremetal deployment'))
        parser.add_argument('--network-ports',
                            help=_('Enable unprovisioning of network ports'),
                            default=False,
                            action="store_true")
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)" % parsed_args)

        with open(parsed_args.input, 'r') as fp:
            roles = yaml.safe_load(fp)

        with oooutils.TempDirs() as tmp:
            unprovision_confirm = os.path.join(tmp, 'unprovision_confirm.json')

            if not parsed_args.yes:
                oooutils.run_ansible_playbook(
                    playbook='cli-overcloud-node-unprovision.yaml',
                    inventory='localhost,',
                    workdir=tmp,
                    playbook_dir=constants.ANSIBLE_TRIPLEO_PLAYBOOKS,
                    verbosity=oooutils.playbook_verbosity(self=self),
                    extra_vars={
                        "stack_name": parsed_args.stack,
                        "baremetal_deployment": roles,
                        "all": parsed_args.all,
                        "prompt": True,
                        "unprovision_confirm": unprovision_confirm,
                        "manage_network_ports": parsed_args.network_ports,
                    }
                )
                with open(unprovision_confirm) as f:
                    to_unprovision = json.load(f)
                    if not to_unprovision:
                        print('Nothing to unprovision, exiting')
                        return
                    self._print_nodes(to_unprovision)

                confirm = oooutils.prompt_user_for_confirmation(
                    message=_("Are you sure you want to unprovision these %s "
                              "nodes [y/N]? ") % parsed_args.stack,
                    logger=self.log)
                if not confirm:
                    raise oscexc.CommandError("Action not confirmed, exiting.")

            oooutils.run_ansible_playbook(
                playbook='cli-overcloud-node-unprovision.yaml',
                inventory='localhost,',
                workdir=tmp,
                playbook_dir=constants.ANSIBLE_TRIPLEO_PLAYBOOKS,
                verbosity=oooutils.playbook_verbosity(self=self),
                extra_vars={
                    "stack_name": parsed_args.stack,
                    "baremetal_deployment": roles,
                    "all": parsed_args.all,
                    "prompt": False,
                    "manage_network_ports": parsed_args.network_ports,
                }
            )

        print('Unprovision complete')

    def _print_nodes(self, nodes):
        TableArgs = collections.namedtuple(
            'TableArgs', 'print_empty max_width fit_width')
        args = TableArgs(print_empty=True, max_width=-1, fit_width=True)
        nodes_data = [(i.get('hostname', ''),
                       i.get('name', ''),
                       i.get('id', '')) for i in nodes]

        sys.stdout.write('\n')
        formatter = table.TableFormatter()
        formatter.emit_list(
            column_names=['hostname', 'name', 'id'],
            data=nodes_data,
            stdout=sys.stdout,
            parsed_args=args
        )
